import sys
import signal
import yoton


if sys.version_info[0] < 3:
    import thread
else:
    import _thread as thread


class PyzoKernelControl(yoton.RepChannel):
    """This is a RepChannel object that runs a thread to control the kernel from the IDE."""

    def interrupt(self, command=None, signum=signal.SIGINT):
        """Interrupt the main thread.

        This does not work if the main thread is running extension code.

        Note that on POSIX we can send an OS INT signal, which is faster
        and maybe more effective in some situations.

        signum must be signal.SIGINT for kernels running Python < 3.10
        """
        if signum == signal.SIGINT:
            thread.interrupt_main()
        else:
            if sys.version_info < (3, 10):
                print("This feature is not implemented for Python versions < 3.10!")
                return
            thread.interrupt_main(signum)

    def terminate(self, command=None):
        """terminate()

        Ask the kernel to terminate by closing the stdin.

        """
        sys.stdin._channel.close()



import urllib.request
import urllib.parse

from pyzo.qt import QtCore, QtWidgets

imported_qtwebkit = True
try:
    from pyzo.qt import QtWebKit
except ImportError:
    imported_qtwebkit = False

import pyzo

tool_name = pyzo.translate("pyzoWebBrowser", "Web browser")
tool_summary = "A very simple web browser."


default_bookmarks = [
    "docs.python.org",
    "scipy.org",
    "doc.qt.nokia.com/4.5/",
    "pyzo.org",
]


class WebView(QtWidgets.QTextBrowser):
    """Inherit the webview class to implement zooming using
    the mouse wheel.
    """

    loadStarted = QtCore.Signal()
    loadFinished = QtCore.Signal(bool)

    def __init__(self, parent):
        super().__init__(parent)

        # Current url
        self._url = ""
        self._history = []
        self._history2 = []

        # Connect
        self.anchorClicked.connect(self.load)

    def wheelEvent(self, event):
        # Zooming does not work for this widget
        if (
            QtCore.Qt.KeyboardModifier.ControlModifier
            & QtWidgets.qApp.keyboardModifiers()
        ):
            self.parent().wheelEvent(event)
        else:
            super().wheelEvent(event)

    def url(self):
        return self._url

    def _getUrlParts(self):
        r = urllib.parse.urlparse(self._url)
        base = r.scheme + "://" + r.netloc
        return base, r.path, r.fragment

    #
    #     def loadCss(self, urls=[]):
    #         urls.append('http://docs.python.org/_static/default.css')
    #         urls.append('http://docs.python.org/_static/pygments.css')
    #         text = ''
    #         for url in urls:
    #             tmp = urllib.request.urlopen(url).read().decode('utf-8')
    #             text += '\n' + tmp
    #         self.document().setDefaultStyleSheet(text)

    def back(self):
        # Get url and update forward history
        url = self._history.pop()
        self._history2.append(self._url)

        # Go there
        url = self._load(url)

    def forward(self):
        if not self._history2:
            return

        # Get url and update forward history
        url = self._history2.pop()
        self._history.append(self._url)

        # Go there
        url = self._load(url)

    def load(self, url):
        # Clear forward history
        self._history2 = []

        # Store current url in history
        while self._url in self._history:
            self._history.remove(self._url)
        self._history.append(self._url)

        # Load
        url = self._load(url)

    def _load(self, url):
        """Convert url and load page, returns new url."""
        # Make url a string
        if isinstance(url, QtCore.QUrl):
            url = str(url.toString())

        # Compose relative url to absolute
        if url.startswith("#"):
            base, path, frag = self._getUrlParts()
            url = base + path + url
        elif "//" not in url:
            base, path, frag = self._getUrlParts()
            url = base + "/" + url.lstrip("/")

        # Try loading
        self.loadStarted.emit()
        self._url = url
        try:
            # print('URL:', url)
            text = urllib.request.urlopen(url).read().decode("utf-8")
            self.setHtml(text)
            self.loadFinished.emit(True)
        except Exception as err:
            self.setHtml(str(err))
            self.loadFinished.emit(False)

        # Set
        return url


class PyzoWebBrowser(QtWidgets.QFrame):
    """The main window, containing buttons, address bar and
    browser widget.
    """

    def __init__(self, parent):
        super().__init__(parent)

        # Init config
        toolId = self.__class__.__name__.lower()
        self._config = pyzo.config.tools[toolId]
        if not hasattr(self._config, "zoomFactor"):
            self._config.zoomFactor = 1.0
        if not hasattr(self._config, "bookMarks"):
            self._config.bookMarks = []
        for item in default_bookmarks:
            if item not in self._config.bookMarks:
                self._config.bookMarks.append(item)

        # Get style object (for icons)
        style = QtWidgets.QApplication.style()

        # Create some buttons
        self._back = QtWidgets.QToolButton(self)
        self._back.setIcon(style.standardIcon(style.StandardPixmap.SP_ArrowBack))
        self._back.setIconSize(QtCore.QSize(16, 16))
        #
        self._forward = QtWidgets.QToolButton(self)
        self._forward.setIcon(style.standardIcon(style.StandardPixmap.SP_ArrowForward))
        self._forward.setIconSize(QtCore.QSize(16, 16))

        # Create address bar
        # self._address = QtWidgets.QLineEdit(self)
        self._address = QtWidgets.QComboBox(self)
        self._address.setEditable(True)
        self._address.setInsertPolicy(self._address.InsertPolicy.NoInsert)
        #
        for a in self._config.bookMarks:
            self._address.addItem(a)
        self._address.setEditText("")

        # Create web view
        if imported_qtwebkit:
            self._view = QtWebKit.QWebView(self)
        else:
            self._view = WebView(self)
        #
        #         self._view.setZoomFactor(self._config.zoomFactor)
        #         settings = self._view.settings()
        #         settings.setAttribute(settings.JavascriptEnabled, True)
        #         settings.setAttribute(settings.PluginsEnabled, True)

        # Layout
        self._sizer1 = QtWidgets.QVBoxLayout(self)
        self._sizer2 = QtWidgets.QHBoxLayout()
        #
        self._sizer2.addWidget(self._back, 0)
        self._sizer2.addWidget(self._forward, 0)
        self._sizer2.addWidget(self._address, 1)
        #
        self._sizer1.addLayout(self._sizer2, 0)
        self._sizer1.addWidget(self._view, 1)
        #
        self._sizer1.setSpacing(2)
        # set margins
        margin = pyzo.config.view.widgetMargin
        self._sizer1.setContentsMargins(margin, margin, margin, margin)

        self.setLayout(self._sizer1)

        # Bind signals
        self._back.clicked.connect(self.onBack)
        self._forward.clicked.connect(self.onForward)
        self._address.lineEdit().returnPressed.connect(self.go)
        self._address.activated.connect(self.go)
        self._view.loadFinished.connect(self.onLoadEnd)
        self._view.loadStarted.connect(self.onLoadStart)

        # Start
        self._view.show()
        self.go("https://docs.python.org")

    def parseAddress(self, address):
        if not address.startswith("http"):
            address = "https://" + address
        return address  # QtCore.QUrl(address, QtCore.QUrl.TolerantMode)

    def go(self, address=None):
        if not isinstance(address, str):
            address = self._address.currentText()
        self._view.load(self.parseAddress(address))

    def onLoadStart(self):
        self._address.setEditText("<loading>")

    def onLoadEnd(self, ok):
        if ok:
            # url = self._view.url()
            # address = str(url.toString())
            if imported_qtwebkit:
                address = self._view.url().toString()
            else:
                address = self._view.url()
        else:
            address = "<could not load page>"
        self._address.setEditText(str(address))

    def onBack(self):
        self._view.back()

    def onForward(self):
        self._view.forward()

    def wheelEvent(self, event):
        if (
            QtCore.Qt.KeyboardModifier.ControlModifier
            & QtWidgets.qApp.keyboardModifiers()
        ):
            # Get amount of scrolling
            degrees = event.delta() / 8.0
            steps = degrees / 15.0
            # Set factor
            factor = self._view.zoomFactor() + steps / 10.0
            if factor < 0.25:
                factor = 0.25
            if factor > 4.0:
                factor = 4.0
            # Store and apply
            self._config.zoomFactor = factor
        #             self._view.setZoomFactor(factor)
        else:
            super().wheelEvent(event)



"""
Simple example Pyzo tool for running custom commands via some buttons

Place this file "myRunner.py" in one of the following two folders:
a) the "tools" folder inside the Pyzo source directory
b) the "tools" folder inside the Pyzo userdata directory

To obtain the Pyzo source directory and Pyzo userdata directory,
see the "About Pyzo" dialog ("Help -> About Pyzo" in the menu).

You can even interactively modify the object of this class in Pyzo's Logger tool:
    obj = pyzo.toolManager.getTool('myrunner')
    obj._btn1.setText('xyz')
And you will also see the output of normal print commands in the Logger tools.

After changing this file, select "Tools -> Reload tools" in Pyzo's menu.

For more information, look at: https://github.com/pyzo/pyzo/blob/main/pyzo/tools/__init__.py
"""

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
from pyzo import translate

tool_name = translate("myRunner", "My Runner")
tool_summary = "Run custom actions."

Qt = QtCore.Qt

class MyRunner(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        # Layout
        self._layout = QtWidgets.QHBoxLayout(self)
        self.setLayout(self._layout)

        # Widgets
        btn = QtWidgets.QPushButton(icon=pyzo.icons.script, text='abc')
        btn.clicked.connect(lambda: self.onButtonClicked('btn1'))
        btn.setShortcut('Ctrl+Shift+1')
        btn.setToolTip('insert some text in the editor')
        self._layout.addWidget(btn)
        self._btn1 = btn

        btn = QtWidgets.QPushButton(icon=pyzo.icons.monitor, text='def')
        btn.clicked.connect(lambda: self.onButtonClicked('btn2'))
        btn.setShortcut('Ctrl+Shift+2')
        btn.setToolTip('run some command in the shell')
        self._layout.addWidget(btn)
        self._btn2 = btn

        btn = QtWidgets.QPushButton(icon=pyzo.icons.wand, text='test')
        btn.clicked.connect(lambda: self.onButtonClicked('btn3'))
        btn.setShortcut('Ctrl+Shift+3')
        btn.setToolTip('show a message box')
        self._layout.addWidget(btn)
        self._btn3 = btn

        btn = QtWidgets.QPushButton(icon=pyzo.icons.wrench_orange, text='')
        btn.clicked.connect(lambda: self.onButtonClicked('btn4'))
        btn.setShortcut('Ctrl+Shift+4')
        btn.setToolTip('show a message box')
        self._layout.addWidget(btn)
        self._btn4 = btn

        self._layout.addStretch()

    def onButtonClicked(self, s):
        print(self.__class__.__name__, 'onButtonClicked', s)
        if s == 'btn1':
            # manipulate the text of the current editor
            editor = pyzo.editors.getCurrentEditor()
            editor.insertPlainText(
                'This text will be inserted\nat the cursor pos of the current editor\n'
            )
        elif s == 'btn2':
            # run a command in the current Python shell
            shell = pyzo.shells.getCurrentShell()
            shell.executeCode('myvar = 123\nprint(myvar + 4)', '<myrunner>')
        else:
            QtWidgets.QMessageBox.information(self, tool_name,'button "{}" clicked'.format(s))

    def close(self):
        pass  # do some clean-up here if necessary



"""pyzowizard module

Implements a wizard to help new users get familiar with pyzo.

"""

import os
import re

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets
from pyzo import translate

from pyzo.util._locale import LANGUAGES, LANGUAGE_SYNONYMS, setLanguage


def retranslate(t):
    """To allow retranslating after selecting the language."""
    if hasattr(t, "original"):
        return translate("wizard", t.original)
    else:
        return t


class PyzoWizard(QtWidgets.QWizard):
    def __init__(self, parent):
        super().__init__(parent)

        # Set some appearance stuff
        self.setMinimumSize(600, 500)
        self.setWindowTitle(translate("wizard", "Getting started with Pyzo"))
        self.setWizardStyle(self.WizardStyle.ModernStyle)
        self.setButtonText(self.WizardButton.CancelButton, "Stop")

        # Set logo
        pm = QtGui.QPixmap()
        pm.load(os.path.join(pyzo.pyzoDir, "resources", "appicons", "pyzologo48.png"))
        self.setPixmap(self.WizardPixmap.LogoPixmap, pm)

        # Define pages
        klasses = [
            IntroWizardPage,
            TwocomponentsWizardPage,
            EditorWizardPage,
            ShellWizardPage1,
            ShellWizardPage2,
            RuncodeWizardPage1,
            RuncodeWizardPage2,
            DebuggingWizardPage1,
            DebuggingWizardPage2,
            ToolsWizardPage1,
            ToolsWizardPage2,
            AdvAutoCompWizardPage,
            MiscellaneousWizardPage,
            FinalPage,
        ]

        # Create pages
        self._n = len(klasses)
        for i, klass in enumerate(klasses):
            self.addPage(klass(self, i))

    def show(self, startPage=None):
        """Show the wizard. If startPage is given, open the Wizard at
        that page. startPage can be an integer or a string that matches
        the classname of a page.
        """
        super().show()

        # Check startpage
        if isinstance(startPage, int):
            pass
        elif isinstance(startPage, str):
            for i in range(self._n):
                page = self.page(i)
                if page.__class__.__name__.lower() == startPage.lower():
                    startPage = i
                    break
            else:
                print("Pyzo wizard: Could not find start page: {!r}".format(startPage))
                startPage = None
        elif startPage is not None:
            print("Pyzo wizard: invalid start page: {!r}".format(startPage))
            startPage = None

        # Go to start page
        if startPage is not None:
            for i in range(startPage):
                self.next()


class BasePyzoWizardPage(QtWidgets.QWizardPage):
    _prefix = translate("wizard", "Step")

    _title = "dummy title"
    _descriptions = []
    _image_filename = ""

    def __init__(self, parent, i):
        super().__init__(parent)
        self._i = i

        # Create label for description
        self._text_label = QtWidgets.QLabel(self)
        self._text_label.setTextFormat(QtCore.Qt.TextFormat.RichText)
        self._text_label.setWordWrap(True)
        self._text_label.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
        )

        # Create label for image
        self._comicLabel = QtWidgets.QLabel(self)
        pm = QtGui.QPixmap()
        if "logo" in self._image_filename:
            pm.load(
                os.path.join(
                    pyzo.pyzoDir, "resources", "appicons", self._image_filename
                )
            )
        elif self._image_filename:
            pm.load(
                os.path.join(pyzo.pyzoDir, "resources", "images", self._image_filename)
            )
        self._comicLabel.setPixmap(pm)
        AF = QtCore.Qt.AlignmentFlag
        self._comicLabel.setAlignment(AF.AlignHCenter | AF.AlignVCenter)

        # Layout
        theLayout = QtWidgets.QVBoxLayout(self)
        self.setLayout(theLayout)
        #
        theLayout.addWidget(self._text_label)
        theLayout.addStretch()
        theLayout.addWidget(self._comicLabel)
        theLayout.addStretch()

    def initializePage(self):
        # Get prefix
        i = self._i
        n = self.wizard()._n - 2  # Don't count the first and last page
        prefix = ""
        if i and i <= n:
            prefix = retranslate(self._prefix) + " {}/{}: ".format(i, n)

        # Set title
        self.setTitle(prefix + retranslate(self._title))

        # Parse description
        # Two description units are separated with BR tags
        # Emphasis on words is set to italic tags.
        lines = []
        descriptions = [retranslate(d).strip() for d in self._descriptions]
        for description in descriptions:
            for line in description.splitlines():
                line = line.strip()
                line = re.sub(r"\*(.+?)\*", r"<b>\1</b>", line)
                lines.append(line)
            lines.append("<br /><br />")
        lines = lines[:-1]

        # Set description
        self._text_label.setText("\n".join(lines))


class IntroWizardPage(BasePyzoWizardPage):
    _title = translate("wizard", "Welcome to the Interactive Editor for Python!")
    _image_filename = "pyzologo128.png"
    _descriptions = [
        translate(
            "wizard",
            """This wizard helps you get familiarized with the workings of Pyzo.""",
        ),
        translate(
            "wizard",
            """Pyzo is a cross-platform Python IDE
        focused on *interactivity* and *introspection*, which makes it
        very suitable for scientific computing. Its practical design
        is aimed at *simplicity* and *efficiency*.""",
        ),
    ]

    def __init__(self, parent, i):
        super().__init__(parent, i)

        # Create label and checkbox
        t1 = translate("wizard", "This wizard can be opened using 'Help > Pyzo wizard'")
        # t2 = translate('wizard', "Show this wizard on startup")
        self._label_info = QtWidgets.QLabel(t1, self)
        # self._check_show = QtWidgets.QCheckBox(t2, self)
        # self._check_show.stateChanged.connect(self._setNewUser)

        # Create language switcher
        self._langLabel = QtWidgets.QLabel(translate("wizard", "Select language"), self)
        #
        self._langBox = QtWidgets.QComboBox(self)
        self._langBox.setEditable(False)
        # Fill
        index, theIndex = -1, -1
        cur = pyzo.config.settings.language
        for lang in sorted(LANGUAGES):
            index += 1
            self._langBox.addItem(lang)
            if lang == LANGUAGE_SYNONYMS.get(cur, cur):
                theIndex = index
        # Set current index
        if theIndex >= 0:
            self._langBox.setCurrentIndex(theIndex)
        # Bind signal
        self._langBox.activated.connect(self.onLanguageChange)

        # Init check state
        # if pyzo.config.state.newUser:
        #    self._check_show.setCheckState(QtCore.Qt.Checked)

        # Create sublayout
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._langLabel, 0)
        layout.addWidget(self._langBox, 0)
        layout.addStretch(2)
        self.layout().addLayout(layout)

        # Add to layout
        self.layout().addSpacing(10)
        self.layout().addWidget(self._label_info)
        # self.layout().addWidget(self._check_show)

    def _setNewUser(self, newUser):
        newUser = bool(newUser)
        self._label_info.setHidden(newUser)
        pyzo.config.state.newUser = newUser

    def onLanguageChange(self):
        languageName = self._langBox.currentText()
        if pyzo.config.settings.language == languageName:
            return
        # Save new language
        pyzo.config.settings.language = languageName
        setLanguage(pyzo.config.settings.language)
        # Notify user
        text = translate(
            "wizard",
            """
        The language has been changed for this wizard.
        Pyzo needs to restart for the change to take effect application-wide.
        """,
        )
        m = QtWidgets.QMessageBox(self)
        m.setWindowTitle(translate("wizard", "Language changed"))
        m.setText(text)
        m.setIcon(m.Icon.Information)
        m.exec()

        # Get props of current wizard
        geo = self.wizard().geometry()
        parent = self.wizard().parent()
        # Close ourself!
        self.wizard().close()
        # Start new one
        w = PyzoWizard(parent)
        w.setGeometry(geo)
        w.show()


class TwocomponentsWizardPage(BasePyzoWizardPage):
    _title = translate("wizard", "Pyzo consists of two main components")
    _image_filename = "pyzo_two_components.png"
    _descriptions = [
        translate("wizard", "You can execute commands directly in the *shell*,"),
        translate("wizard", "or you can write code in the *editor* and execute that."),
    ]


class EditorWizardPage(BasePyzoWizardPage):
    _title = translate("wizard", "The editor is where you write your code")
    _image_filename = "pyzo_editor.png"
    _descriptions = [
        translate(
            "wizard",
            """In the *editor*, each open file is represented as a tab. By
        right-clicking on a tab, files can be run, saved, closed, etc.""",
        ),
        translate(
            "wizard",
            """The right mouse button also enables one to make a file the
        *main file* of a project. This file can be recognized by its star
        symbol, and it enables running the file more easily.""",
        ),
    ]


class ShellWizardPage1(BasePyzoWizardPage):
    _title = translate("wizard", "The shell is where your code gets executed")
    _image_filename = "pyzo_shell1.png"
    _descriptions = [
        translate(
            "wizard",
            """When Pyzo starts, a default *shell* is created. You can add more
        shells that run simultaneously, and which may be of different
        Python versions.""",
        ),
        translate(
            "wizard",
            """Shells run in a sub-process, such that when it is busy, Pyzo
        itself stays responsive, allowing you to keep coding and even
        run code in another shell.""",
        ),
    ]


class ShellWizardPage2(BasePyzoWizardPage):
    _title = translate("wizard", "Configuring shells")
    _image_filename = "pyzo_shell2.png"
    _descriptions = [
        translate(
            "wizard",
            """Pyzo can integrate the event loop of five different *GUI toolkits*,
        thus enabling interactive plotting with e.g. Visvis or Matplotlib.""",
        ),
        translate(
            "wizard",
            """Via 'Shell > Edit shell configurations', you can edit and add
        *shell configurations*. This allows you to for example select the
        initial directory, or use a custom Pythonpath.""",
        ),
    ]


class RuncodeWizardPage1(BasePyzoWizardPage):
    _title = translate("wizard", "Running code")
    _image_filename = "pyzo_run1.png"
    _descriptions = [
        translate(
            "wizard",
            "Pyzo supports several ways to run source code in the editor. (see the 'Run' menu).",
        ),
        translate(
            "wizard",
            """*Execute selection:* if there is no selected text, the current line
        is executed; if the selection is on a single line, the selection
        is evaluated; if the selection spans multiple lines, Pyzo will
        run the (complete) selected lines.""",
        ),
        translate(
            "wizard",
            """*Execute line and print result:* execute the editor's current line
        in the shell and print the result
        <br />
        e.g.: line "<code>y = myfunc(x)</code>" will be executed as "<code>y = myfunc(x); y</code>" """,
        ),
        translate(
            "wizard",
            "*Execute cell:* a cell is everything between two lines starting with '##'.",
        ),
        translate("wizard", "*Execute file:* run all the code in the current file."),
        translate(
            "wizard",
            "*Execute main file:* run the code in the current main file.",
        ),
    ]


class RuncodeWizardPage2(BasePyzoWizardPage):
    _title = translate("wizard", "Interactive mode vs running as script")
    _image_filename = ""
    _descriptions = [
        translate(
            "wizard",
            """You can run the current file or the main file normally, or as a script.
        When run as script, the file is saved first. Then the shell is restarted to provide
        a clean environment before executing the file.
        The shell is also initialized differently so that it closely resembles a normal
        script execution.""",
        ),
        translate(
            "wizard",
            """In *interactive mode*, <code>sys.path[0]</code> is an empty string (i.e. the current dir),
        <code>sys.argv</code> is set to <code>['']</code> and <code>__file__</code> is not defined.""",
        ),
        translate(
            "wizard",
            """As a substitute for the undefined <code>__file__</code>, you can insert and run
            something like""",
        ),
        """<code>import os, inspect<br />
        __file__ = inspect.getfile(inspect.currentframe())<br />
        __this_dir__ = os.path.abspath(os.path.dirname(__file__))
        </code>""",
        translate(
            "wizard",
            """to get it.""",
        ),
        translate(
            "wizard",
            """In *script mode*, <code>__file__</code> and <code>sys.argv[0]</code> are set
            to the script's filename, <code>sys.path[0]</code> and the working dir are set
            to the directory containing the script.""",
        ),
    ]


class DebuggingWizardPage1(BasePyzoWizardPage):
    _title = translate("wizard", "Debugging code and execution control (1/2)")
    _image_filename = ""
    _descriptions = [
        translate(
            "wizard",
            """To stop the execution of code (in an infinite loop or when calculations
        take longer than expected, use the flash symbol, or select "Interrupt" from the
        "Shell" menu. This will raise an exception, and you can postmortem-debug after
        that.""",
        ),
        translate(
            "wizard",
            """Via *Postmortem* debugging, you can inspect the values of objects after
        code execution was stopped by an unhandled exception. By switching stack frames,
        you can see local variables in different layers of nested function calls.
        Continuing execution is not possible in postmortem, though.""",
        ),
        translate(
            "wizard",
            """The *Pause* debug action (in the "Shell" menu or shell toolbar) makes it
        possible to pause execution and activate the debug mode, and then continue, stop
        or pause again the execution.""",
        ),
        translate(
            "wizard",
            """While in *debug mode*, there are many possibilities:
        <ul>
        <li>objects can be inspected and modified</li>
        <li>stack frames can be switched (to go to different levels of nested function calls)
        <li>custom code can be executed in the shell (also whole scripts)</li>
        <li>code execution can be stepped</li>
        <li>the pointer to the next line to execute can be changed via the "Debug jump" action,
        for example to skip lines or repeat previously executed lines (some limitations
        apply, though)</li>
        </ul>
        When continuing execution (and thus leaving debug mode), breakpoints will be updated
        (see next page).""",
        ),
    ]


class DebuggingWizardPage2(BasePyzoWizardPage):
    _title = translate("wizard", "Debugging code and execution control (2/2)")
    _image_filename = ""
    _descriptions = [
        translate(
            "wizard",
            """While the debug mode is active, the gui event loop of the Python interpreter
        in the shell is normally not periodically called anymore. This means that, for example,
        matplotlib figures will be frozen as if the program would have crashed.
        By setting the environment variable <code>PYZO_PROCESS_EVENTS_WHILE_DEBUGGING=1</code>
        in the "environ" field of the shell configuration, the gui event loop will be processed
        even while debug mode is active. Matplotlib figures can then also be interacted with
        normally while code execution is paused.""",
        ),
        translate(
            "wizard",
            """To (repeatedly) pause at specific code lines and enter debug mode, set a
        *breakpoint* there, via "Toggle breakpoint" from the "Edit" menu or by clicking in
        the bar to the right of the line number. Only a full circle is a breakpoint, whereas
        a half circle represents a disabled breakpoint that can be used as a navigation point
        (via "Jump to previous/next breakpoint" from the menu).
        Breakpoints are only updated (i.e. synchronized with the debugger) when the user
        gives the command to execute code or when continuing execution while in debug mode.
        When executing code that has code parts still running via the event loop after
        execution has finished, for example GUI applications with PySide6, a simple way
        to update breakpoints is to just press RETURN in the idle command prompt of the shell.
        Another way to update breakpoints would be to pause execution and continue it.
        Note that, when there are breakpoints set (and synchronized), execution becomes much
        slower than without breakpoints.""",
        ),
        translate(
            "wizard",
            """Instead of a breakpoint or pause action, code execution can also be paused
        by executing the function *<code>breakpoint()</code>* in the Python code. This will
        enter debug mode the same way. Compared to normal breakpoints, there is no
        performance penalty, but with the disadvantage, that these hard-coded breakpoint
        functions cannot be disabled while the script is executed.""",
        ),
    ]


class ToolsWizardPage1(BasePyzoWizardPage):
    _title = translate("wizard", "Tools for your convenience")
    _image_filename = "pyzo_tools1.png"
    _descriptions = [
        translate(
            "wizard",
            """Via the *Tools menu*, one can select which tools to use. The tools can
        be positioned in any way you want, and can also be un-docked.""",
        ),
        translate(
            "wizard",
            """Note that the tools system is designed such that it's easy to
        create your own tools. Look at the example tool for more information,
        or use one of the existing tools as a more advanced example.""",
        ),
        os.path.join(pyzo.pyzoDir, "resources", "code_examples", "myRunner.py"),
        os.path.join(pyzo.pyzoDir, "tools"),
        "https://github.com/pyzo/pyzo/blob/main/pyzo/resources/code_examples/myRunner.py",
        "https://github.com/pyzo/pyzo/blob/main/pyzo/tools/",
    ]


class ToolsWizardPage2(BasePyzoWizardPage):
    _title = translate("wizard", "Recommended tools")
    _image_filename = "pyzo_tools2.png"
    _descriptions = [
        translate("wizard", """We especially recommend the following tools:"""),
        translate(
            "wizard",
            """The *Source structure* tool gives an outline of the source code.""",
        ),
        translate(
            "wizard",
            """The *File browser* tool helps keeping an overview of all files
            in a directory and allows for searching text across multiple files.
            To manage your bookmarks, click the star icon.""",
        ),
        translate(
            "wizard",
            """The *Interactive help* tool shows you the documentation of objects,
            automatically when browsing the autocompletion list,
            or manually by selecting an expression and clicking "Help" in the context menu.""",
        ),
        translate(
            "wizard",
            """The *Editor list* tool is convenient when having more than just
            a few editor tabs open at the same time.""",
        ),
    ]


class AdvAutoCompWizardPage(BasePyzoWizardPage):
    _title = translate("wizard", "Advanced Autocompletion and Calltips")
    _image_filename = ""
    _descriptions = [
        translate(
            "wizard",
            """Auto completion for attributes and calltips can also be used for complex
        expressions that include function calls and indexing.<br />
        <br />
        For example:<br />
        <br />
        Type "<code>locals().</code>" (with the dot at the end) and press Ctrl+Space. This will evaluate
        <code>locals()</code> and assign the result to <code>__pyzo__autocomp</code> in the
        local scope. Then auto-completion continues and displays the list widget with the
        attributes.<br />
        <br />
        Type "<code>locals()[</code>" (with the open bracket at the end) and press Ctrl+Space.
        This will evaluate <code>locals()</code> and assign the result to
        <code>__pyzo__autocomp</code> in the local scope. Then auto-completion continues
        and displays the list widget with the keys.<br />
        <br />
        Type "<code>locals().get(</code>" (with the open parenthesis at the end) and press
        Ctrl+Space. This will evaluate <code>locals()</code> and assign the result to
        <code>__pyzo__calltip</code> in the local scope. Then auto-completion continues
        and displays the calltip for <code>dict.get(...)</code>.<br />
        <br />
        The Ctrl+Space hot key can also be pressed while having already entered some characters
        of the attribute, key or arguments, in case the autocompletion list widget resp. calltip
        should be shown again after having cancelled it.""",
        ),
    ]


class MiscellaneousWizardPage(BasePyzoWizardPage):
    _title = translate("wizard", "Miscellaneous tips and tricks")
    _image_filename = ""
    _descriptions = [
        translate(
            "wizard",
            """Press keys Ctrl+Shift+ArrowUp/Down to move the selected lines up/down
        in the editor.<br />
        <br />
        Use magic commands to manage packets, e.g. "pip list" or "pip install numpy".<br />
        <br />
        When there is a printed stack trace in Python shell (usually the red error text),
        double click the *filepath* to open the file at the specific line number.<br />
        <br />
        When dealing with performance critical code, avoid printing too much to the shell
        because stream outputs directed to the Pyzo IDE will slow down the execution.<br />
        <br />
        You can double-click the line number to get an input field for jumping to a
        specific line.<br />
        <br />
        In Pyzo's "About" dialog (Help -> About Pyzo) you can see the folder where Pyzo
        stores the settings.<br />
        <br />
        Pyzo can be configured to be portable, for example to run it from a USB pen-drive
        with encapsulated settings. To enable this, rename the folder "_settings" in Pyzo's
        application folder to "settings".<br />
        <br />
        You can directly modify Pyzo's source code even in the binary distribution because
        the binary is just a Python-Interpreter that runs the Python source code in the
        installation directory. See the "About Pyzo" dialog for the code location.""",
        ),
    ]


class FinalPage(BasePyzoWizardPage):
    _title = translate("wizard", "Get coding!")
    _image_filename = "pyzologo128.png"
    _descriptions = [
        translate(
            "wizard",
            """This concludes the Pyzo wizard. Now, get coding and have fun!""",
        ),
    ]


# def smooth_images():
#     """ This was used to create the images from their raw versions.
#     """
#
#     import os
#     import visvis as vv
#     import scipy as sp
#     import scipy.ndimage
#     for fname in os.listdir('images'):
#         im = vv.imread(os.path.join('images', fname))
#         for i in range(im.shape[2]):
#             im[:,:,i] = sp.ndimage.gaussian_filter(im[:,:,i], 0.7)
#         #fname = fname.replace('.png', '.jpg')
#         print(fname)
#         vv.imwrite(fname, im[::2,::2,:])


if __name__ == "__main__":
    w = PyzoWizard(None)
    w.show()



"""Subpackage parsers

This subpackage contains all the syntax parsers for the
different languages.

"""

""" CREATING PARSERS

Making a parser requires these things:
  * Place a module in the parsers directory, which has a name
    ending in "_parser.py"
  * In the module implement one or more classes that inherit
    from ..parsers.Parser (or a derived class), and
    implement the parseLine method.
  * The module should import all the tokens in whiches to use
    from ..parsers.tokens. New tokens can also be
    defined by subclassing one of the token classes.
  * In codeeditor/parsers/__init__.py, add the new module to the
    list of imported parsers.

"""

import sys

from . import tokens


class BlockState:
    """BlockState(state=0, info=None)

    The blockstate object should be used by parsers to
    return the block state of the processed line.

    This would typically be the last token in a line, but this might also
    be before the last token in a line. Even multiple BlockState objects
    can be present in a line, in which case the last one is considered valid.
    """

    isToken = False

    def __init__(self, state=0, info=None):
        self._state = int(state)
        self._info = info

    @property
    def state(self):
        """The integer value representing the block state."""
        return self._state

    @property
    def info(self):
        """Get the information corresponding to the block."""
        return self._info


# Base parser class (needs to be defined before importing parser modules)
class Parser:
    """Base parser class.
    All parsers should inherit from this class.
    This base class generates a 'TextToken' for each line
    """

    _extensions = []
    _shebangKeywords = []
    _keywords = []

    @classmethod
    def getParserName(cls):
        name = cls.__name__
        if name.endswith("Parser") and len(name) >= 6:
            name = name[:-6].lower()
        return name

    @classmethod
    def disambiguate(cls, text):
        return cls.getParserName()

    def parseLine(self, line, previousState=0):
        """The method that should be implemented by the parser. The
        previousState argument can be used to determine how
        the previous block ended (e.g. for multiline comments). It
        is an integer, the meaning of which is only known to the
        specific parser.

        This method should reaturn a list of token instances. The
        last token can be a BlockState to specify the previousState
        for the next block.
        """

        return [tokens.TextToken(line, 0, len(line))]

    def name(self):
        """Get the name of the parser."""
        name = self.__class__.__name__.lower()
        if name.endswith("parser"):
            name = name[:-6]
        return name

    def __repr__(self):
        """String representation of the parser."""
        return '<Parser for "{}">'.format(self.name())

    def keywords(self):
        """Get a list of keywords valid for this parser."""
        return self._keywords[:]

    def filenameExtensions(self):
        """Get a list of filename extensions for which this parser is appropriate."""
        return ["." + e.lstrip(".").lower() for e in self._extensions]

    def shebangKeywords(self):
        """Get a list of shebang keywords for which this parser is appropriate."""
        return self._shebangKeywords.copy()

    def getStyleElementDescriptions(cls):
        """This method returns a list of the StyleElementDescription
        instances used by this parser.
        """
        descriptions = {}
        for token in cls.getUsedTokens(cls):
            descriptions[token.description.key] = token.description

        return list(descriptions.values())

    def getUsedTokens(self):
        """Get a a list of token instances used by this parser."""

        # Get module object of the parser
        try:
            mod = sys.modules[self.__module__]
        except KeyError:
            return []

        # Get token classes from module
        tokenClasses = []
        for name in mod.__dict__:
            member = mod.__dict__[name]
            if isinstance(member, type) and issubclass(member, tokens.Token):
                if member is not tokens.Token:
                    tokenClasses.append(member)

        # Return as instances
        return [t() for t in tokenClasses]

    def _isTodoItem(self, text):
        """Get whether the given text (which should be a comment) represents
        a todo item. Todo items start with "todo", "2do" or "fixme",
        optionally with a colon at the end.
        """
        # Get first word
        word = text.lstrip().split(" ", 1)[0].rstrip(":")
        # Test
        if word.lower() in ["todo", "2do", "fixme"]:
            return True
        else:
            return False


## Import parsers statically
# We could load the parser dynamically from the source files in the
# directory, but this takes quite some effort to get right when apps
# are frozen. This is doable (I do it in Visvis) but it requires the
# user to specify the parser modules by hand when freezing an app.
#
# In summary: it takes a lot of trouble, which can be avoided by just
# listing all parsers here.

from . import python_parser  # noqa
from . import cython_parser  # noqa
from . import c_parser  # noqa
from . import s_expr_parser  # noqa



# -*- coding: utf-8 -*-
# This example demonstrates simple pub sub.
#
# This time in event driven mode. This example only works locally, as
# we cannot start two event loops :)

## ========== One end

import yoton

verbosity = 0

# Create a context and a pub channel
ct1 = yoton.Context(verbose=verbosity)
pub = yoton.PubChannel(ct1, "foo")

# Connect and turn duplicator on
ct1.bind("publichost:test")


## ========== Other end

import yoton

verbosity = 0

# Create a context and a sub channel
ct2 = yoton.Context(verbose=verbosity)
sub = yoton.SubChannel(ct2, "foo")

# Connect, set channel to event driven mode
ct2.connect("publichost:test")


# Create message handler
def message_handler():
    message = sub.recv(False)
    if message:
        print(message)
        if message.lower() == "stop":
            yoton.stop_event_loop()


# Bind handler to a timer
timer = yoton.Timer(0.1, False)
timer.bind(message_handler)
timer.start()


# Send messages
yoton.call_later(pub.send, 8, "stop")
yoton.call_later(pub.send, 2, "2 seconds")
yoton.call_later(pub.send, 4, "4 seconds")
yoton.call_later(pub.send, 6, "almost done")

# Enter event loop
yoton.start_event_loop()



"""Module editor

Defines the PyzoEditor class which is used to edit documents.
This module/class also implements all the relatively low level
file loading/saving /reloading stuff.

"""

import os
import sys
import re
import codecs

from pyzo.qt import QtCore, QtGui, QtWidgets

from pyzo.codeeditor import Manager
from pyzo.core.menu import EditorContextMenu
from pyzo.core.baseTextCtrl import BaseTextCtrl, normalizePath
from pyzo.core.pyzoLogging import print
import pyzo


"""
Handling of encodings and byte-order-marks
==========================================

When loading text from a file, the encoding is determined as follows, similar to PEP 263:
    1) If there is a known BOM, the encoding is defined by the BOM.
    2) Otherwise, if there is a magic comment with a known encoding name,
       then this encoding name will be used.  e.g.: "# -*- coding: latin-1 -*-"
    3) Otherwise utf-8 will be used.

    If decoding fails, other encoding candidates from 2) or 3) will be tried as well.
    If decoding still fails, ascii with backslashreplace will be used, a warning will
    be displayed, and the document will be marked as modified after loading.

And when saving text to a file:
    1) If there is a magic comment with a known encoding name in the unsaved text,
       that encoding will be used for saving the file.
    2) Otherwise utf-8 encoding will be used.

    If there was a BOM_UTF8 present when originally loading the file, that BOM will
    only be written to the file if the encoding for saving is utf-8 or utf-8-sig.
    Other BOMs, such as BOM_UTF16_LE, will always be discarded when writing a file.

"""


def getEncodingFromMagic(text):
    """reads the encoding name from the file header
    according to https://peps.python.org/pep-0263/
    returns 'utf-8' if no or an unknown encoding name was found
    """
    encoding = "utf-8"
    for line in text.splitlines()[:2]:
        mo = re.match(r"^[ \t\f]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+)", line)
        if mo:
            try:
                encoding = codecs.lookup(mo[1]).name
            except LookupError:
                pass
            break
        if line.split("#", 1)[0].lstrip():
            break  # line is not just whitespace and/or comment
    return encoding


def getEncodingFromBom(databytes):
    """detects the byte order mark and returns the encoding name or None"""
    bom2enc = {
        codecs.BOM_UTF8: "utf-8-sig",
        codecs.BOM_UTF16_BE: "utf-16-be",
        codecs.BOM_UTF16_LE: "utf-16-le",
        codecs.BOM_UTF32_BE: "utf-32-be",
        codecs.BOM_UTF32_LE: "utf-32-le",
    }
    for bom, encoding in bom2enc.items():
        if databytes[:4].startswith(bom):
            break
    else:
        encoding = None
    return encoding


# Set default line ending (if not set)
if not pyzo.config.settings.defaultLineEndings:
    if sys.platform.startswith("win"):
        pyzo.config.settings.defaultLineEndings = "CRLF"
    else:
        pyzo.config.settings.defaultLineEndings = "LF"


def determineLineEnding(text):
    """Get the line ending style used in the text.
    \n, \r, \r\n,
    The EOLmode is determined by counting the occurrences of each
    line ending...
    """
    text = text[:32768]  # Limit search for large files

    # test line ending by counting the occurrence of each
    c_win = text.count("\r\n")
    c_mac = text.count("\r") - c_win
    c_lin = text.count("\n") - c_win
    # set the appropriate style
    if c_win > c_mac and c_win > c_lin:
        mode = "\r\n"
    elif c_mac > c_win and c_mac > c_lin:
        mode = "\r"
    else:
        mode = "\n"

    # return
    return mode


def determineIndentation(text):
    return determineIndentationAndTrailingWS(text)[0]


def determineIndentationAndTrailingWS(text):
    """Get the indentation used in this document and whether there is
    any trailing whitespace.
    The text is analyzed to find the most used indentations.
    The result is -1 if tab indents are most common.
    A positive result means spaces are used; the amount
    signifies the amount of spaces per indentation.
    0 is returned if the indentation could not be determined.
    The second return value is the number of lines with trailing ws.
    """

    text = text[:32768]  # Limit search for large files

    # create dictionary of indents, -1 means a tab
    indents = {}
    indents[-1] = 0
    trailing = 0

    lines = text.splitlines()
    lines.insert(0, "")  # so the lines start at 1
    for i, line in enumerate(lines):
        # remove indentation
        lineA = line.lstrip()
        lineB = line.rstrip()
        lineC = lineA.rstrip()
        indent = len(line) - len(lineA)
        if len(lineB) < len(line):
            trailing += 1

        line = lineC
        if line.startswith("#"):
            continue
        else:
            # remove everything after the #
            line = line.split("#", 1)[0].rstrip()
        if not line:
            # continue of no line left
            continue

        # a colon means there will be an indent
        # check the next line (or the one thereafter)
        # and calculate the indentation difference with THIS line.
        if line.endswith(":"):
            if len(lines) > i + 2:
                line2 = lines[i + 1]
                tmp = line2.lstrip()
                if not tmp:
                    line2 = lines[i + 2]
                    tmp = line2.lstrip()
                if tmp:
                    ind2 = len(line2) - len(tmp)
                    ind3 = ind2 - indent
                    if line2.startswith("\t"):
                        indents[-1] += 1
                    elif ind3 > 0:
                        if ind3 not in indents:
                            indents[ind3] = 1
                        indents[ind3] += 1

    # find which was the most common tab width.
    indent, maxvotes = 0, 0
    for nspaces in indents:
        if indents[nspaces] > maxvotes:
            indent, maxvotes = nspaces, indents[nspaces]
    # print("found tabwidth", indent)
    return indent, trailing


# To give each new file a unique name
newFileCounter = 0


def createEditor(parent, filename=None):
    """Tries to load the file given by the filename and
    if succesful, creates an editor instance to put it in,
    which is returned.
    If filename is None, an new/unsaved/temp file is created.
    """

    if filename is None:
        # Increase counter
        global newFileCounter
        newFileCounter += 1

        # Create editor
        editor = PyzoEditor(parent)
        editor.document().setModified(True)
        editor.removeTrailingWS = True

        # Set name
        editor._name = "<tmp {}>".format(newFileCounter)

    else:
        # check and normalize
        if not os.path.isfile(filename):
            raise IOError("File does not exist '{}'.".format(filename))

        # create editor
        editor = PyzoEditor(parent)

        text = editor._loadTextFromFile(filename)

        # store name and filename
        editor._filename = filename
        editor._name = os.path.split(filename)[1]

    if editor._filename:
        editor._modifyTime = os.path.getmtime(editor._filename)

    # Set parser
    if editor._filename:
        ext = os.path.splitext(editor._filename)[1]
        parser = Manager.suggestParser(ext, text)
        editor.setParser(parser)
    else:
        # todo: rename style -> parser
        editor.setParser(pyzo.config.settings.defaultStyle)

    # return
    return editor


class PyzoEditor(BaseTextCtrl):
    # called when dirty changed or filename changed, etc
    somethingChanged = QtCore.Signal()

    def __repr__(self):
        return "<{} - {}, {}>".format(self.__class__.__qualname__, id(self), self.name)

    def __init__(self, parent, **kwds):
        super().__init__(parent, showLineNumbers=True, **kwds)

        # Init filename and name
        self._filename = ""
        self._name = "<TMP>"

        # View settings
        # TODO: self.setViewWrapSymbols(view.showWrapSymbols)
        self.setShowLineEndings(pyzo.config.view.showLineEndings)
        self.setShowIndentationGuides(pyzo.config.view.showIndentationGuides)
        #
        self.setWrap(bool(pyzo.config.view.wrap))
        self.setHighlightCurrentLine(pyzo.config.view.highlightCurrentLine)
        self.setLongLineIndicatorPosition(pyzo.config.view.edgeColumn)
        # TODO: self.setFolding( int(view.codeFolding)*5 )
        # bracematch is set in baseTextCtrl, since it also applies to shells
        # dito for zoom and tabWidth

        # Set line endings to default
        self.lineEndings = pyzo.config.settings.defaultLineEndings

        # Do not use utf-8-sig byte order mark as a default
        self.useBom = False

        # Modification time to test file change
        self._modifyTime = 0
        self._binaryDataFromFile = None  # the binary data as it was loaded or saved

        self.modificationChanged.connect(self._onModificationChanged)

        # To see whether the doc has changed to update the parser.
        self.textChanged.connect(self._onModified)

        # This timer is used to hide the marker that shows which code is executed
        self._showRunCursorTimer = QtCore.QTimer()

        # Add context menu
        self._menu = EditorContextMenu(self)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._onContextMenu)

        # Update status bar
        self.cursorPositionChanged.connect(self._updateStatusBar)

    def closeEvent(self, event):
        # free some memory to mitigate memory leak losses
        del self._binaryDataFromFile

    ## Properties

    @property
    def name(self):
        return self._name

    @property
    def filename(self):
        return self._filename

    @property
    def lineEndings(self):
        """Line-endings style of this file.

        Setter accepts machine-readable (e.g. '\r') and human-readable (e.g. 'CR') input
        """
        return self._lineEndings

    @lineEndings.setter
    def lineEndings(self, value):
        if value in ("\r", "\n", "\r\n"):
            self._lineEndings = value
            return
        try:
            self._lineEndings = {"CR": "\r", "LF": "\n", "CRLF": "\r\n"}[value]
        except KeyError:
            raise ValueError("Invalid line endings style {!r}".format(value))

    @property
    def lineEndingsHumanReadable(self):
        """Current line-endings style, human readable (e.g. 'CR')"""
        return {"\r": "CR", "\n": "LF", "\r\n": "CRLF"}[self.lineEndings]

    ##

    def _loadTextFromFile(self, filepath):
        with open(filepath, "rb") as f:
            bb = f.read()

        # determine encoding from byte order mark and PEP 263 magic comment
        encBom = getEncodingFromBom(bb)
        encMagic = getEncodingFromMagic(bb[:200].decode("ascii", "replace"))

        self.useBom = encBom == "utf-8-sig" and encMagic.startswith("utf-8")

        encMagic2 = encMagic
        if encMagic2 == "utf-8":
            encMagic2 = "utf-8-sig"  # utf-8-sig also works for utf-8 encoded text

        encodingsToTry = []
        if encBom:
            encodingsToTry.append(encBom)
        if encMagic2 not in encodingsToTry:
            encodingsToTry.append(encMagic2)
        if "utf-8-sig" not in encodingsToTry:
            encodingsToTry.append("utf-8-sig")

        ok = False
        for encoding in encodingsToTry:
            try:
                text = bb.decode(encoding)
                ok = True
                break
            except ValueError:
                pass

        if not ok:
            text = bb.decode("ascii", "backslashreplace")

            msg = (
                'Error decoding contents of file "{}". '
                "Already tried {} without success. "
                "--> now using ascii with backslashreplace".format(
                    filepath, encodingsToTry
                )
            )

            print(msg)
            m = QtWidgets.QMessageBox(self)
            m.setWindowTitle("Error while loading text")
            m.setText(msg)
            m.setIcon(m.Icon.Warning)
            m.exec()

        # process line endings
        self.lineEndings = determineLineEnding(text)

        # process indentation and trailing
        indentWidth, trailing = determineIndentationAndTrailingWS(text)
        self.removeTrailingWS = trailing < 10  # not too much for ugly diffs
        if indentWidth == -1:  # Tabs
            self.setIndentWidth(pyzo.config.settings.defaultIndentWidth)
            self.setIndentUsingSpaces(False)
        elif indentWidth:
            self.setIndentWidth(indentWidth)
            self.setIndentUsingSpaces(True)

        # set text
        self.setPlainText(text)
        self.document().setModified(not ok)
        self._binaryDataFromFile = bb
        return text

    def _saveTextToFile(self, filepath):
        text = self.toPlainText().replace("\n", self.lineEndings)
        encoding = getEncodingFromMagic(text[:200])
        if encoding == "utf-8" and self.useBom:
            encoding = "utf-8-sig"

        bb = text.encode(encoding)
        with open(filepath, "wb") as f:
            f.write(bb)

        if not encoding.startswith("utf-8"):
            self.useBom = False

        self._binaryDataFromFile = bb

        print(
            "saved file: {} ({}, {})".format(
                filepath, encoding, self.lineEndingsHumanReadable
            )
        )

    ##

    def justifyText(self):
        """Overloaded version of justifyText to make it use our
        configurable justificationwidth.
        """
        super().justifyText(pyzo.config.settings.justificationWidth)

    def showRunCursor(self, cursor):
        """Momentarily highlight a piece of code to show that this is being executed"""

        extraSelection = QtWidgets.QTextEdit.ExtraSelection()
        extraSelection.cursor = cursor
        extraSelection.format.setBackground(QtCore.Qt.GlobalColor.gray)
        self.setExtraSelections([extraSelection])

        self._showRunCursorTimer.singleShot(200, lambda: self.setExtraSelections([]))

    def id(self):
        """Get an id of this editor. This is the filename, or for tmp files, the name."""
        if self._filename:
            return self._filename
        else:
            return self._name

    def focusInEvent(self, event):
        """Test whether the file has been changed 'behind our back'"""
        pyzo.callLater(self._checkAndHandleExternalFileModifications)
        return super().focusInEvent(event)

    def _checkAndHandleExternalFileModifications(self):
        if not self._checkFileModified():
            return

        if self.document().isModified():
            # ask user
            dlg = QtWidgets.QMessageBox(self)
            dlg.setWindowTitle("File was changed")
            dlg.setText(
                "File has also been modified outside of the editor:\n" + self._filename
            )
            dlg.setInformativeText("Do you want to reload?")
            btnReload = dlg.addButton(
                "Reload", QtWidgets.QMessageBox.ButtonRole.AcceptRole
            )
            dlg.addButton(
                "Keep this version", QtWidgets.QMessageBox.ButtonRole.RejectRole
            )
            dlg.setDefaultButton(btnReload)

            # get user's decision
            dlg.exec()
            doReload = dlg.clickedButton() == btnReload
        else:
            if pyzo.config.advanced.autoReloadFilesInEditor:
                doReload = True
                print("auto-reloading file", self._filename)
            else:
                # ask user
                dlg = QtWidgets.QMessageBox(self)
                dlg.setWindowTitle("File was changed")
                dlg.setText(
                    "File has been modified outside of the editor:\n" + self._filename
                )
                dlg.setInformativeText("Do you want to reload?")
                btnAlwaysReload = dlg.addButton(
                    "Always reload", QtWidgets.QMessageBox.ButtonRole.NoRole
                )
                btnReloadThisTime = dlg.addButton(
                    "Reload", QtWidgets.QMessageBox.ButtonRole.AcceptRole
                )
                dlg.addButton(
                    "Keep this version", QtWidgets.QMessageBox.ButtonRole.RejectRole
                )
                dlg.setDefaultButton(btnReloadThisTime)

                # get user's decision
                dlg.exec()
                if dlg.clickedButton() == btnReloadThisTime:
                    doReload = True
                elif dlg.clickedButton() == btnAlwaysReload:
                    doReload = True
                    pyzo.config.advanced.autoReloadFilesInEditor = 1

                    QtWidgets.QMessageBox.information(
                        self,
                        "Always reload",
                        'Files that are marked as "unchanged" in the editor will now '
                        "always be reloaded automatically. You can change this setting "
                        '"autoReloadFilesInEditor" in the "Advanced settings" dialog.',
                    )
                else:
                    doReload = False

        if doReload:
            self.reload()
        else:
            self.document().setModified(True)

    def _checkFileModified(self):
        """Test if the file was externally changed since it was last loaded/saved by Pyzo.
        Returns True if it was changed, False if not changed or unknown.
        """

        filepath = self._filename
        if not os.path.isfile(filepath):
            # file is deleted from the outside
            return False

        # test the modification time...
        mtime = os.path.getmtime(filepath)

        if mtime == self._modifyTime:
            return False

        # whatever the user will decide, we will reset the modified time
        self._modifyTime = mtime

        try:
            with open(filepath, "rb") as f:
                bb = f.read()
            if bb == self._binaryDataFromFile:
                print(
                    "modification time of file {} changed, but content is the same".format(
                        filepath
                    )
                )
                return False
        except Exception:
            pass

        return True

    def _onModificationChanged(self, changed):
        """Handler for the modificationChanged signal. Emit somethingChanged
        for the editorStack to update the modification notice."""
        self.somethingChanged.emit()

    def _onModified(self):
        pyzo.parser.parseThis(self)

    def _onContextMenu(self, point):
        cur = self.textCursor()
        curFromEvent = self.cursorForPosition(point)
        moveTextCursor = True
        if cur.hasSelection():
            i1 = cur.selectionStart()
            i2 = cur.selectionEnd()
            if i1 <= curFromEvent.position() <= i2:
                moveTextCursor = False
        if moveTextCursor:
            self.setTextCursor(curFromEvent)
        # the offset is to prevent accidental auto-clicking
        self._menu.popup(self.mapToGlobal(point) + QtCore.QPoint(0, 3))

    def _updateStatusBar(self):
        editor = pyzo.editors.getCurrentEditor()
        sb = pyzo.main.statusBar()
        sb.updateCursorInfo(editor)

    def dragMoveEvent(self, event):
        """Otherwise cursor can get stuck.
        https://bitbucket.org/iep-project/iep/issue/252
        https://qt-project.org/forums/viewthread/3180
        """
        if event.mimeData().hasUrls():
            event.acceptProposedAction()
        else:
            super().dropEvent(event)

    def dropEvent(self, event):
        """Drop files in the list."""
        if event.mimeData().hasUrls():
            # file: let the editorstack do the work.
            pyzo.editors.dropEvent(event)
        else:
            # text: act normal
            super().dropEvent(event)

    def showEvent(self, event=None):
        """Capture show event to change title."""
        # Act normally
        if event:
            super().showEvent(event)

        # Make parser update
        pyzo.parser.parseThis(self)

    def setTitleInMainWindow(self):
        """set the title text in the main window to show filename."""
        name, path = self._name, self._filename
        pyzo.main.setMainTitle(path or name)

    def save(self, filename=None):
        """Save the file. No checking is done."""

        # get filename
        if filename is None:
            filename = self._filename
        if not filename:
            raise ValueError("No filename specified, and no filename known.")

        # Test whether it was changed without us knowing. If so, don't save now.
        if self._checkFileModified():
            # ask user
            dlg = QtWidgets.QMessageBox(self)
            dlg.setWindowTitle("File was changed")
            dlg.setText(
                "File has also been modified outside of the Pyzo editor:\n"
                + self._filename
            )
            dlg.setInformativeText("Do you want to reload?")
            btnOverwrite = dlg.addButton(
                "Overwrite file", QtWidgets.QMessageBox.ButtonRole.AcceptRole
            )
            btnCancel = dlg.addButton(
                "Cancel saving", QtWidgets.QMessageBox.ButtonRole.RejectRole
            )
            dlg.setDefaultButton(btnCancel)

            # get user's decision
            dlg.exec()
            if dlg.clickedButton() != btnOverwrite:
                # cancel saving
                self.document().setModified(True)
                return

        # Remove whitespace in a single undo-able action
        if (
            self.removeTrailingWS
            or pyzo.config.settings.removeTrailingWhitespaceWhenSaving
        ):
            # Original cursor to put state back at the end
            oricursor = self.textCursor()
            # Screen cursor to select document
            screenCursor = self.textCursor()
            screenCursor.movePosition(screenCursor.MoveOperation.Start)
            screenCursor.movePosition(
                screenCursor.MoveOperation.End,
                screenCursor.MoveMode.KeepAnchor,
            )
            # Cursor for doing the editor
            editCursor = self.textCursor()
            # Go!
            editCursor.beginEditBlock()

            try:
                editCursor.setPosition(screenCursor.selectionStart())
                editCursor.movePosition(editCursor.MoveOperation.StartOfBlock)
                while (
                    editCursor.position() < screenCursor.selectionEnd()
                    or editCursor.position() <= screenCursor.selectionStart()
                ):
                    editCursor.movePosition(editCursor.MoveOperation.StartOfBlock)
                    editCursor.movePosition(
                        editCursor.MoveOperation.EndOfBlock,
                        editCursor.MoveMode.KeepAnchor,
                    )
                    text1 = editCursor.selectedText()
                    text2 = text1.rstrip()
                    if len(text1) != len(text2):
                        editCursor.insertText(text2)
                    if not editCursor.block().next().isValid():
                        break
                    editCursor.movePosition(editCursor.MoveOperation.NextBlock)
            finally:
                # Setting the textcursor might scroll to another place
                # To prevent this, we store and then restore the position
                # of the vertical scrollbar
                sb = self.verticalScrollBar()
                scrollbarPos = sb.value()
                self.setTextCursor(oricursor)
                editCursor.endEditBlock()
                sb.setValue(scrollbarPos)

        self._saveTextToFile(filename)

        # Update stats
        self._filename = normalizePath(filename)
        self._name = os.path.split(self._filename)[1]
        self.document().setModified(False)
        self._modifyTime = os.path.getmtime(self._filename)

        # update title (in case of a rename)
        # Saving could be triggered for a non-active editor, e.g. when running the
        # main file as a script.
        if self is pyzo.editors.getCurrentEditor():
            self.setTitleInMainWindow()

        # allow item to update its texts (no need: onModifiedChanged does this)
        # self.somethingChanged.emit()

    def saveCopy(self, filepath):
        """Creates a backup of the current editor's contents. No checking is done."""
        self._saveTextToFile(filepath)

    def reload(self):
        """Reload text using the self._filename.
        We do not have a load method; we first try to load the file
        and only when we succeed create an editor to show it in...
        This method is only for reloading in case the file was changed
        outside of the editor."""

        # We can only load if the filename is known
        if not self._filename:
            return
        filename = self._filename

        # Remember where we are
        cursor = self.textCursor()
        linenr = cursor.blockNumber() + 1

        self._loadTextFromFile(filename)

        # Go where we were (approximately)
        self.gotoLine(linenr)

        # Trigger update of indentation and line ending in the "File" menu
        pyzo.editors.onCurrentChanged()

    def _expandSelectionToWholeBlocks(self):
        """expands the selection of the current text cursor to whole blocks

        The selection will be expanded so that it starts at the beginning of the block
        and so that it ends at the beginning of the next block, including the new-block
        special character.
        If there is no new block after the end of the expanded selection, which can only
        occur at the end of the document, then "endsAtNextBlock" will be set to False.

        If no text is selected, the whole block will be selected.

        return value:
            cursor, endsAtNextBlock
        """
        cursor = self.textCursor()
        # Find start and end of selection
        start = cursor.selectionStart()
        end = cursor.selectionEnd()
        # Expand selection to the start of the first block
        cursor.setPosition(start)
        cursor.movePosition(cursor.MoveOperation.StartOfBlock)

        # Expand selection to the start of the next block if there is no selection at all
        # or if there is at least one character selected in the last selected block.
        cursor.setPosition(end, cursor.MoveMode.KeepAnchor)
        if not cursor.atBlockStart() or not cursor.hasSelection():
            endsAtNextBlock = cursor.movePosition(
                cursor.MoveOperation.NextBlock,
                cursor.MoveMode.KeepAnchor,
            )
            if not endsAtNextBlock:
                # Could not move to the beginning of the next block because the end of the
                # document was reached. Therefore move till the end of the document.
                cursor.movePosition(
                    cursor.MoveOperation.End, cursor.MoveMode.KeepAnchor
                )
        else:
            # There is selected text and the cursor is at the beginning of the block.
            endsAtNextBlock = True
        return cursor, endsAtNextBlock

    def deleteLines(self):
        cursor, endsAtNextBlock = self._expandSelectionToWholeBlocks()
        cursor.removeSelectedText()

    def duplicateLines(self):
        cursor, endsAtNextBlock = self._expandSelectionToWholeBlocks()
        text = cursor.selectedText()
        if not endsAtNextBlock:
            text += "\u2029"
        cursor.setPosition(cursor.selectionStart())
        cursor.insertText(text)

    def commentCode(self):
        """
        Comment the lines that are currently selected
        """
        indents = []
        indentChar = " " if self.indentUsingSpaces() else "\t"

        def getIndent(cursor):
            text = cursor.block().text().rstrip()
            if text:
                indents.append(len(text) - len(text.lstrip()))

        def commentBlock(cursor):
            blockText = cursor.block().text()
            numMissingIndentChars = minindent - (
                len(blockText) - len(blockText.lstrip(indentChar))
            )
            if numMissingIndentChars > 0:
                # Prevent setPosition from leaving bounds of the current block
                # if there are too few indent characters (e.g. an empty line)
                cursor.insertText(indentChar * numMissingIndentChars)
            cursor.setPosition(cursor.block().position() + minindent)
            cursor.insertText("# ")

        self.doForSelectedBlocks(getIndent)
        minindent = min(indents, default=0)
        self.doForSelectedBlocks(commentBlock)

    def uncommentCode(self):
        """Uncomment the lines that are currently selected"""
        # TODO: this should not be applied to lines that are part of a multi-line string

        # Define the uncomment function to be applied to all blocks
        def uncommentBlock(cursor):
            """Find the first # on the line; if there is just whitespace before it,
            remove the # and if it is followed by a space remove the space, too
            """
            text = cursor.block().text()
            commentStart = text.find("#")
            if commentStart == -1:
                return  # No comment on this line
            if text[:commentStart].strip() != "":
                return  # Text before the #
            # Move the cursor to the beginning of the comment
            cursor.setPosition(cursor.block().position() + commentStart)
            cursor.deleteChar()
            if text[commentStart:].startswith("# "):
                cursor.deleteChar()

        # Apply this function to all blocks
        self.doForSelectedBlocks(uncommentBlock)

    def toggleCommentCode(self):
        def toggleComment():
            """Toggles comments for the seclected text in editor, most of the code is
            taken from commentCode and uncommentCode
            """
            text_block = []

            def getBlocks(cursor):
                text = cursor.block().text()
                text_block.append(text)

            def commentBlock(cursor):
                cursor.setPosition(cursor.block().position())
                cursor.insertText("# ")

            def uncommentBlock(cursor):
                """Find the first # on the line; if there is just whitespace before it,
                remove the # and if it is followed by a space remove the space, too
                """
                cursor.setPosition(cursor.block().position())
                cursor.deleteChar()
                cursor.deleteChar()

            self.doForSelectedBlocks(getBlocks)
            commented = [item for item in text_block if item.startswith("# ")]
            if len(commented) == len(text_block):
                self.doForSelectedBlocks(uncommentBlock)
            else:
                self.doForSelectedBlocks(commentBlock)

        toggleComment()

    def gotoDef(self):
        """Goto the definition for the word under the cursor"""

        # Get name of object to go to
        cursor = self.textCursor()
        if not cursor.hasSelection():
            cursor.select(cursor.SelectionType.WordUnderCursor)
        word = cursor.selection().toPlainText()

        # Send the open command to the shell
        s = pyzo.shells.getCurrentShell()
        if s is not None:
            if word and word.isidentifier():
                s.executeCommand("open {}\n".format(word))
            else:
                s.write("Invalid identifier {!r}\n".format(word))

    ## Introspection processing and autocompletion methods

    def processCallTip(self, cto):
        """Processes a calltip request using a CallTipObject instance."""
        # Try using buffer first
        if cto.tryUsingBuffer():
            return

        # Try obtaining calltip from the source
        sig = pyzo.parser.getFictiveSignature(cto.name, self, True)
        if sig:
            # Done
            cto.finish(sig)
        else:
            # Try the shell
            shell = pyzo.shells.getCurrentShell()
            if shell:
                shell.processCallTip(cto)

    def processAutoComp(self, aco):
        """Processes an autocomp request using an AutoCompObject instance."""

        # Try using buffer first
        if aco.tryUsingBuffer():
            return

        # Get normal fictive namespace
        fictiveNS = pyzo.parser.getFictiveNameSpace(self)
        fictiveNS = set(fictiveNS)

        snipTool = pyzo.toolManager.getTool("pyzosnippets")

        # Add names
        if not aco.name:
            # "root" names
            aco.addNames(fictiveNS)
            # imports
            importNames, importLines = pyzo.parser.getFictiveImports(self)
            aco.addNames(importNames)
            if snipTool:
                aco.addNames(["SNIP"])
        else:
            if snipTool:
                names = snipTool.getAutocompleteNames(aco.name)
                if names is not None:
                    # aco.name is "SNIP" or "SNIP.something"
                    aco.addNames(names)
                    aco.finish()
                    return

            # Prepare list of class names to check out
            classNames = [aco.name]
            handleSelf = True
            # Unroll supers
            while classNames:
                className = classNames.pop(0)
                if not className:
                    continue
                if handleSelf or (className in fictiveNS):
                    # Only the self list (only first iter)
                    fictiveClass = pyzo.parser.getFictiveClass(
                        className, self, handleSelf
                    )
                    handleSelf = False
                    if fictiveClass:
                        aco.addNames(fictiveClass.members)
                        classNames.extend(fictiveClass.supers)
                        if className in classNames:
                            # avoid infinite loop if the parent class is the same as the
                            # current class, see https://github.com/pyzo/pyzo/issues/944
                            break
                else:
                    break

        # If there's a shell, let it finish the autocompletion
        shell = pyzo.shells.getCurrentShell()
        if shell:
            shell.processAutoComp(aco)
        else:
            # Otherwise we finish it ourselves
            aco.finish()

    def _onAutocompleteFinished(self, completionPerformed):
        if completionPerformed:
            snipTool = pyzo.toolManager.getTool("pyzosnippets")
            if snipTool:
                snipTool.applyAutoCompletion(self.textCursor())

        super()._onAutocompleteFinished(completionPerformed)

    def _onAutocompleteSelectionChanged(self, name):
        snipTool = pyzo.toolManager.getTool("pyzosnippets")
        if snipTool:
            fullName = self._getFullAutocompleteName(name)
            if fullName is not None and snipTool.isSnipName(fullName):
                snipTool.showHelpForSnip(fullName)
                return
        super()._onAutocompleteSelectionChanged(name)


if __name__ == "__main__":
    # Do some stubbing to run this module as a unit separate from pyzo
    # TODO: untangle pyzo from this module where possible
    class DummyParser:
        def parseThis(self, x):
            pass

    pyzo.parser = DummyParser()
    EditorContextMenu = QtWidgets.QMenu
    app = QtWidgets.QApplication([])
    win = PyzoEditor(None)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+C"), win).activated.connect(win.copy)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+X"), win).activated.connect(win.cut)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+V"), win).activated.connect(win.paste)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Shift+V"), win).activated.connect(
        win.pasteAndSelect
    )
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Z"), win).activated.connect(win.undo)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Y"), win).activated.connect(win.redo)

    tmp = "foo(bar)\nfor bar in range(5):\n  print bar\n"
    tmp += "\nclass aap:\n  def monkey(self):\n    pass\n\n"
    win.setPlainText(tmp)
    win.show()
    app.exec()



"""Module pyzokernel.interpreter

Implements the Pyzo interpreter.

Notes on IPython
----------------
We integrate IPython via the IPython.core.interactiveshell.InteractiveShell.
  * The namespace is set to __main__
  * We call its run_cell method to execute code
  * Debugging/breakpoints are "enabled using the pre_run_code_hook
  * Debugging occurs in our own debugger
  * GUI integration is all handled by pyzo
  * We need special prompts for IPython input



"""

import os
import re
import sys
import time
import logging
import platform
import struct
import shlex
from codeop import CommandCompiler
import traceback
import keyword
import inspect  # noqa - Must be in this namespace
import bdb
import linecache
import signal
import tokenize

import yoton
from pyzokernel import guiintegration, printDirect
from pyzokernel.magic import Magician
from pyzokernel.debug import Debugger

# Init last traceback information
sys.last_type = None
sys.last_value = None
sys.last_traceback = None

# Set Python version and get some names
PYTHON_VERSION = sys.version_info[0]
if PYTHON_VERSION < 3:
    ustr = unicode  # noqa
    bstr = str
    input = raw_input  # noqa
else:
    ustr = str
    bstr = bytes


file_linenr_pattern = re.compile(r"(?:\.py|<tmp \d+>)(\+\d+)['\",]+ line (\d+)")


class PS1:
    """Dynamic prompt for PS1. Show IPython prompt if available, and
    show current stack frame when debugging.
    """

    def __init__(self, pyzo):
        self._pyzo = pyzo

    def __str__(self):
        if self._pyzo._dbFrames:
            # When debugging, show where we are, do not use IPython prompt
            preamble = "(" + self._pyzo._dbFrameName + ")"
            return "\n\x1b[0;32m%s>>>\x1b[0m " % preamble
        elif self._pyzo._ipython:
            # IPython prompt
            return "\n\x1b[0;32mIn [\x1b[1;32m%i\x1b[0;32m]:\x1b[0m " % (
                self._pyzo._ipython.execution_count
            )
            # return 'In [%i]: ' % (self._ipython.execution_count)
        else:
            # Normal Python prompt
            return "\n\x1b[0;32m>>>\x1b[0m "


class PS2:
    """Dynamic prompt for PS2."""

    def __init__(self, pyzo):
        self._pyzo = pyzo

    def __str__(self):
        if self._pyzo._dbFrames:
            # When debugging, show where we are, do not use IPython prompt
            preamble = "(" + self._pyzo._dbFrameName + ")"
            return "\x1b[0;32m%s...\x1b[0m " % preamble
        elif self._pyzo._ipython:
            # Dots ala IPython
            nspaces = len(str(self._pyzo._ipython.execution_count)) + 2
            return "\x1b[0;32m%s...:\x1b[0m " % (nspaces * " ")
        else:
            # Just dots
            return "\x1b[0;32m...\x1b[0m "


class PyzoInterpreter:
    """PyzoInterpreter

    The pyzo interpreter is the part that makes the pyzo kernel interactive.
    It executes code, integrates the GUI toolkit, parses magic commands, etc.
    The pyzo interpreter has been designed to emulate the standard interactive
    Python console as much as possible, but with a lot of extra goodies.

    There is one instance of this class, stored at sys._pyzoInterpreter and
    at the __pyzo__ variable in the global namespace.

    The global instance has a couple of interesting attributes:
      * context: the yoton Context instance at the kernel (has all channels)
      * introspector: the introspector instance (a subclassed yoton.RepChannel)
      * magician: the object that handles the magic commands
      * guiApp: a wrapper for the integrated GUI application

    """

    # Simular working as code.InteractiveConsole. Some code was copied, but
    # the following things are changed:
    # - prompts are printed in the err stream, like the default interpreter does
    # - uses an asynchronous read using the yoton interface
    # - support for hijacking GUI toolkits
    # - can run large pieces of code
    # - support post mortem debugging
    # - support for magic commands

    def __init__(self, globals, locals, filename="<console>"):
        # Init variables for locals and globals (globals only for debugging)
        self.globals = globals
        self.locals = locals

        # Store filename
        self._filename = filename

        # Store ref of globals and locals so that we can restore them
        # after a debugging session with different frames.
        self._main_globals = globals
        self._main_locals = locals

        # Flag to ignore sys exit, to allow running some scripts
        # interactively, even if they call sys.exit.
        self.ignore_sys_exit = False

        # Information for debugging. If self._dbFrames, we're in debug mode
        # _dbFrameIndex starts from 1
        self._dbFrames = []
        self._dbFrameIndex = 0
        self._dbFrameName = ""

        # Init datase to store source code that we execute
        self._codeCollection = ExecutedSourceCollection()

        # Init buffer to deal with multi-line command in the shell
        self._buffer = []

        # Create compiler
        if sys.platform.startswith("java"):
            import compiler

            self._compile = compiler.compile  # or 'exec' does not work
        else:
            self._compile = CommandCompiler()

        # Instantiate magician and tracer
        self.magician = Magician()
        self.debugger = Debugger()

        # To keep track of whether to send a new prompt, and whether more
        # code is expected.
        self.more = 0
        self.newPrompt = True
        self._oldPromptString = None

        # Code and script to run on first iteration
        self._codeToRunOnStartup = None
        self._scriptToRunOnStartup = None

        # Remove "THIS" directory from the PYTHONPATH
        # to prevent unwanted imports. Same for pyzokernel dir
        thisPath = os.getcwd()
        for p in [thisPath, os.path.join(thisPath, "pyzokernel")]:
            while p in sys.path:
                sys.path.remove(p)

    def run(self):
        """Run (start the mainloop)

        Here we enter the main loop, which is provided by the guiApp.
        This event loop calls process_commands on a regular basis.

        We may also enter the debug interaction loop, either from a
        request for post-mortem debugging, or *during* execution by
        means of a breakpoint. When in this debug-loop, the guiApp event
        loop lays still, but the debug-loop does call process-commands
        for user interaction.

        When the user wants to quit, SystemExit is raised (one way or
        another). This is detected in process_commands and the exception
        instance is stored in self._exitException. Then the debug-loop
        is stopped if necessary, and the guiApp is told to stop its event
        loop.

        And that brings us back here, where we exit using in order of
        preference: self._exitException, the exception with which the
        event loop was exited (if any), or a new exception.

        """

        # Prepare
        self._prepare()
        self._exitException = None

        # Enter main
        try:
            self.guiApp.run(self.process_commands)
        except SystemExit:
            # Set self._exitException if it is not set yet
            type, value, tb = sys.exc_info()
            del tb
            if self._exitException is None:
                self._exitException = value
        finally:
            self._cleanup()

        # Exit
        if self._exitException is None:
            self._exitException = SystemExit()
        raise self._exitException

    def _cleanup(self):
        # restore prompts
        if self._original_ps1 is not None:
            # sys.ps1 and sys.ps2 are only defined in interactive mode
            sys.ps1 = self._original_ps1
            sys.ps2 = self._original_ps2

    def _prepare(self):
        """Prepare for running the main loop.
        Here we do some initialization like obtaining the startup info,
        creating the GUI application wrapper, etc.
        """

        # Reset debug status
        self.debugger.writestatus()

        # Get startup info (get a copy, or setting the new version won't trigger!)
        while self.context._stat_startup.recv() is None:
            time.sleep(0.02)
        self.startup_info = startup_info = self.context._stat_startup.recv().copy()

        # Set startup info (with additional info)
        if sys.platform.startswith("java"):
            import __builtin__ as builtins  # Jython
        else:
            builtins = __builtins__
        if not isinstance(builtins, dict):
            builtins = builtins.__dict__
        startup_info["builtins"] = list(builtins)
        startup_info["version"] = tuple(sys.version_info)
        startup_info["keywords"] = keyword.kwlist
        startup_info["softkeywords"] = getattr(keyword, "softkwlist", [])
        # Update startup info, we update again at the end of this method
        self.context._stat_startup.send(startup_info.copy())

        # Prepare the Python environment
        self._prepare_environment(startup_info)

        # Run startup code (before loading GUI toolkit or IPython
        self._run_startup_code(startup_info)

        # Write Python banner (to stdout)
        thename = "Python"
        if sys.version_info[0] == 2:
            thename = "Legacy Python"
        if "__pypy__" in sys.builtin_module_names:
            thename = "Pypy"
        if sys.platform.startswith("java"):
            thename = "Jython"
            # Jython cannot do struct.calcsize("P")
            import java.lang

            real_plat = java.lang.System.getProperty("os.name").lower()
            plat = "%s/%s" % (sys.platform, real_plat)
        elif sys.platform.startswith("win"):
            NBITS = 8 * struct.calcsize("P")
            plat = "Windows (%i bits)" % NBITS
        else:
            NBITS = 8 * struct.calcsize("P")
            plat = "%s (%i bits)" % (sys.platform, NBITS)
        printDirect(
            "%s %s on %s.\n" % (thename, sys.version.split("[")[0].rstrip(), plat)
        )

        # Integrate GUI
        guiName, guiError = self._integrate_gui(startup_info)

        # Write pyzo part of banner (including what GUI loop is integrated)
        if True:
            pyzoBanner = "This is the Pyzo interpreter"
        if guiError:
            pyzoBanner += ". " + guiError + "\n"
        elif guiName:
            pyzoBanner += " with integrated event loop for "
            pyzoBanner += guiName + ".\n"
        else:
            pyzoBanner += ".\n"
        printDirect(pyzoBanner)

        # Try loading IPython
        if startup_info.get("ipython", "").lower() in ("", "no", "false"):
            self._ipython = None
        else:
            try:
                self._load_ipython()
            except Exception:
                type, value, tb = sys.exc_info()
                del tb
                printDirect("IPython could not be loaded: %s\n" % str(value))
                self._ipython = None
                startup_info["ipython"] = "no"
            if not self._ipython:
                startup_info["ipython"] = "no"

        # Set prompts
        self._original_ps1 = getattr(sys, "ps1", None)
        self._original_ps2 = getattr(sys, "ps2", None)
        sys.ps1 = PS1(self)
        sys.ps2 = PS2(self)

        # Notify about project path
        projectPath = startup_info["projectPath"]
        if projectPath:
            printDirect("Prepending the project path %r to sys.path\n" % projectPath)

        # Write tips message.
        if self._ipython:
            import IPython

            printDirect(
                "\nUsing IPython %s -- An enhanced Interactive Python.\n"
                % IPython.__version__
            )
            printDirect(
                "?         -> Introduction and overview of IPython's features.\n"
                "%quickref -> Quick reference.\n"
                "help      -> Python's own help system.\n"
                "object?   -> Details about 'object', "
                "use 'object??' for extra details.\n"
            )
        else:
            printDirect(
                "Type 'help' for help, " + "type '?' for a list of *magic* commands.\n"
            )

        # Notify the running of the script
        if self._scriptToRunOnStartup:
            printDirect(
                '\x1b[0;33mRunning script: "'
                + self._scriptToRunOnStartup
                + '"\x1b[0m\n'
            )

        # Prevent app nap on OSX 9.2 and up
        # The _nope module is taken from MINRK's appnope package
        if sys.platform == "darwin":

            def parse_version_crudely(version_string):
                """extracts the leading number parts of a version string to a tuple
                e.g.: "123.45ew6.7x.dev8" --> (123, 45, 7)
                """
                return tuple(
                    int(s) for s in re.findall(r"\.(\d+)", "." + version_string)
                )

            parsev = parse_version_crudely
            if parsev(platform.mac_ver()[0]) >= parsev("10.9"):
                from pyzokernel import _nope

                _nope.nope()

        self._add_postmortem_for_logged_exceptions()

        # Setup pausing of running code using SIGFPE
        signal.signal(signal.SIGFPE, self._handle_sigfpe)

        # Update startup info
        self.context._stat_startup.send(startup_info)

    def _add_postmortem_for_logged_exceptions(self):
        # Setup post-mortem debugging via appropriately logged exceptions.
        #
        # The Logger.callHandlers look like a good place to Monkey-patch, as
        # it's certain that the record is going to be handled at this point,
        # and it does not interfere with the rest of the logging mechanics.
        #
        # Adding a handler to the root logger (as we did earlier) seems like
        # an elegant approach, but it gets complicated as to whether Pyzo should
        # take responsibility for printing the message. The fact that a handler
        # is present, will stop the logging module from printing the message,
        # because it considers it handled. We can detect whether our handler
        # is the only one and print if it is, but the log action may be
        # initiated with a "higher-level" logger, so this may cause the message
        # to be printed twice.

        def callHandlers_patched(self, record):
            if record.exc_info:
                sys.last_type, sys.last_value, sys.last_traceback = record.exc_info
            return callHandlers_ori(self, record)

        try:
            callHandlers_ori = logging.Logger.callHandlers
            logging.Logger.callHandlers = callHandlers_patched
        except Exception:
            pass

    def _handle_sigfpe(self, sig, frame):
        self.debugger.set_trace(frame)

    def _prepare_environment(self, startup_info):
        """Prepare the Python environment. There are two possibilities:
        either we run a script or we run interactively.
        """

        # Get whether we should (and can) run as script
        scriptFilename = startup_info["scriptFile"]
        if scriptFilename:
            if not os.path.isfile(scriptFilename):
                printDirect('Invalid script file: "' + scriptFilename + '"\n')
                scriptFilename = None

        # Get project path
        projectPath = startup_info["projectPath"]

        if scriptFilename.endswith(".ipynb"):
            # Run Jupyter notebook
            import notebook.notebookapp

            sys.argv = ["jupyter_notebook", scriptFilename]
            sys.exit(notebook.notebookapp.main())

        elif scriptFilename:
            # RUN AS SCRIPT
            # Set __file__  (note that __name__ is already '__main__')
            self.locals["__file__"] = scriptFilename
            # Set command line arguments
            sys.argv[:] = []
            sys.argv.append(scriptFilename)
            sys.argv.extend(shlex.split(startup_info.get("argv", "")))
            # Insert script directory to path
            theDir = os.path.abspath(os.path.dirname(scriptFilename))
            if theDir not in sys.path:
                sys.path.insert(0, theDir)
            if projectPath is not None:
                sys.path.insert(0, projectPath)
            # Go to script dir
            os.chdir(os.path.dirname(scriptFilename))
            # Run script later
            self._scriptToRunOnStartup = scriptFilename
        else:
            # RUN INTERACTIVELY
            # No __file__ (note that __name__ is already '__main__')
            self.locals.pop("__file__", "")
            # Remove all command line arguments, set first to empty string
            sys.argv[:] = []
            sys.argv.append("")
            sys.argv.extend(shlex.split(startup_info.get("argv", "")))
            # Insert current directory to path
            sys.path.insert(0, "")
            if projectPath:
                sys.path.insert(0, projectPath)
            # Go to start dir
            startDir = startup_info["startDir"]
            if startDir and os.path.isdir(startDir):
                os.chdir(startDir)
            else:
                os.chdir(os.path.expanduser("~"))  # home dir

    def _run_startup_code(self, startup_info):
        """Execute the startup code or script."""

        # Run startup script (if set)
        script = startup_info["startupScript"]
        # Should we use the default startupScript?
        if script == "$PYTHONSTARTUP":
            script = os.environ.get("PYTHONSTARTUP", "")

        if "\n" in script:
            # Run code later or now
            linesBefore = []
            linesAfter = script.splitlines()
            while linesAfter:
                if linesAfter[0].startswith("# AFTER_GUI"):
                    linesAfter.pop(0)
                    break
                linesBefore.append(linesAfter.pop(0))
            scriptBefore = "\n".join(linesBefore)
            self._codeToRunOnStartup = "\n".join(linesAfter)
            if scriptBefore.strip():  # don't trigger when only empty lines
                self.context._stat_interpreter.send("Busy")
                msg = {"source": scriptBefore, "fname": "<startup>", "lineno": 0}
                self.runlargecode(msg, True)
        elif script and os.path.isfile(script):
            # Run script
            self.context._stat_interpreter.send("Busy")
            self.runfile(script)
        else:
            # Nothing to run
            pass

    def _integrate_gui(self, startup_info):
        """Integrate event loop of GUI toolkit (or use pure Python
        event loop).
        """

        self.guiApp = guiintegration.App_nogui()
        self.guiName = guiName = startup_info["gui"].upper()
        guiError = ""

        try:
            if guiName in ["", "NONE"]:
                guiName = ""
            elif guiName == "AUTO":
                for tryName, tryApp in [
                    ("PYSIDE6", guiintegration.App_pyside6),
                    ("PYQT6", guiintegration.App_pyqt6),
                    ("PYSIDE2", guiintegration.App_pyside2),
                    ("PYQT5", guiintegration.App_pyqt5),
                    # ('WX', guiintegration.App_wx),
                    ("ASYNCIO", guiintegration.App_asyncio_new),
                    ("TK", guiintegration.App_tk),
                ]:
                    try:
                        self.guiApp = tryApp()
                    except Exception:
                        continue
                    guiName = tryName
                    break
                else:
                    guiName = ""
            elif guiName == "ASYNCIO":
                self.guiApp = guiintegration.App_asyncio_new()
            elif guiName == "TK":
                self.guiApp = guiintegration.App_tk()
            elif guiName == "WX":
                self.guiApp = guiintegration.App_wx()
            elif guiName == "TORNADO":
                self.guiApp = guiintegration.App_tornado()
            elif guiName == "PYSIDE6":
                self.guiApp = guiintegration.App_pyside6()
            elif guiName == "PYSIDE2":
                self.guiApp = guiintegration.App_pyside2()
            elif guiName in ["PYQT6", "QT6"]:
                self.guiApp = guiintegration.App_pyqt6()
            elif guiName in ["PYQT5", "QT5"]:
                self.guiApp = guiintegration.App_pyqt5()
            elif guiName == "FLTK":
                self.guiApp = guiintegration.App_fltk()
            elif guiName == "GTK":
                self.guiApp = guiintegration.App_gtk()
            else:
                guiError = "Unknown gui: %s" % guiName
                guiName = ""
        except Exception:  # Catch any error
            # Get exception info (we do it using sys.exc_info() because
            # we cannot catch the exception in a version independent way.
            type, value, tb = sys.exc_info()
            del tb
            guiError = "Failed to integrate event loop for %s: %s" % (
                guiName,
                str(value),
            )

        return guiName, guiError

    def _load_ipython(self):
        """Try loading IPython shell. The result is set in self._ipython
        (can be None if IPython not available).
        """

        # Init
        self._ipython = None
        import __main__

        # Try importing IPython
        try:
            import IPython
        except ImportError:
            return

        # Version ok?
        if IPython.version_info < (1,):
            return

        # Create an IPython shell
        from IPython.core.interactiveshell import InteractiveShell

        self._ipython = InteractiveShell(user_module=__main__)

        # Set some hooks / event callbacks
        # Run hook (pre_run_code_hook is depreacted in 2.0)
        pre_run_cell_hook = self.ipython_pre_run_cell_hook
        if IPython.version_info < (2,):
            self._ipython.set_hook("pre_run_code_hook", pre_run_cell_hook)
        else:
            self._ipython.events.register("pre_run_cell", pre_run_cell_hook)
        # Other hooks
        self._ipython.set_hook("editor", self.ipython_editor_hook)
        self._ipython.set_custom_exc((bdb.BdbQuit,), self.dbstop_handler)

        # Some patching
        self._ipython.ask_exit = self.ipython_ask_exit

        # Make output be shown on Windows
        if sys.platform.startswith("win"):
            # IPython wraps std streams just like we do below, but
            # pyreadline adds *another* wrapper, which is where it
            # goes wrong. Here we set it back to bypass pyreadline.

            if IPython.version_info < (
                8,
            ):  # corrects a problem with IOStream from IPython 8.x.x
                from IPython.utils import io

                io.stdin = io.IOStream(sys.stdin)
                io.stdout = io.IOStream(sys.stdout)
                io.stderr = io.IOStream(sys.stderr)

            # Ipython uses msvcrt e.g. for pausing between pages
            # but this does not work in pyzo
            import msvcrt

            msvcrt.getwch = msvcrt.getch = input  # input is deffed above

    def process_commands(self):
        """Do one iteration of processing commands (the REPL)."""
        try:
            self._process_commands()

        except SystemExit:
            # It may be that we should ignore sys exit now...
            if self.ignore_sys_exit:
                self.ignore_sys_exit = False  # Never allow more than once
                return
            # Get and store the exception so we can raise it later
            type, value, tb = sys.exc_info()
            del tb
            self._exitException = value
            # Stop debugger if it is running
            self.debugger.stopinteraction()
            # Exit from interpreter. Exit in the appropriate way
            self.guiApp.quit()  # Is sys.exit() by default

    def _process_commands(self):
        # Run startup code/script inside the loop (only the first time)
        # so that keyboard interrupt will work
        if self._codeToRunOnStartup:
            self.context._stat_interpreter.send("Busy")
            self._codeToRunOnStartup, tmp = None, self._codeToRunOnStartup
            self._runlines(tmp, filename="<startup_after_gui>", symbol="exec")
        if self._scriptToRunOnStartup:
            self.context._stat_interpreter.send("Busy")
            self._scriptToRunOnStartup, tmp = None, self._scriptToRunOnStartup
            self.runfile(tmp)

        # Flush real stdout / stderr
        sys.__stdout__.flush()
        sys.__stderr__.flush()

        if self.introspector._run_mode == 0:
            # Introspection was set up without its own thread or event loop (see start.py).
            # So we have to regularly process introspection requests here.
            self.introspector._process_events_local()

        # Set status and prompt?
        # Prompt is allowed to be an object with __str__ method
        # We also compare prompt strings because stack frame changes
        # caused by a breakpoint while running the gui event loop
        # would get unnotified otherwise.
        promptString = str(sys.ps2 if self.more else sys.ps1)
        if self.newPrompt or promptString != self._oldPromptString:
            self.newPrompt = False
            self._oldPromptString = promptString
            self.context._strm_prompt.send(promptString)

        if True:
            # Determine state. The message is really only send
            # when the state is different. Note that the kernelbroker
            # can also set the state ("Very busy", "Busy", "Dead")
            if self._dbFrames:
                self.context._stat_interpreter.send("Debug")
            elif self.more:
                self.context._stat_interpreter.send("More")
            else:
                self.context._stat_interpreter.send("Ready")
            try:
                cwd = os.getcwd()
            except Exception:
                # On Linux, if the current working directory has been deleted,
                # then a FileNotFoundError Exception is raised.
                # We cannot send an empty string, so we send a "?" to indicate the error.
                cwd = "?"
            self.context._stat_cd.send(cwd)

        # Are we still connected?
        if sys.stdin.closed or not self.context.connection_count:
            # Exit from main loop.
            # This will raise SystemExit and will shut us down in the
            # most appropriate way
            sys.exit()

        # Get channel to take a message from
        ch = yoton.select_sub_channel(
            self.context._ctrl_command, self.context._ctrl_code
        )

        if ch is None:
            pass  # No messages waiting

        elif ch is self.context._ctrl_command:
            # Read command
            line1 = self.context._ctrl_command.recv(False)  # Command
            if line1:
                # Notify what we're doing
                self.context._strm_echo.send(line1)
                self.context._stat_interpreter.send("Busy")
                self.newPrompt = True
                # Convert command
                # (only a few magics are supported if IPython is active)
                line2 = self.magician.convert_command(line1.rstrip("\n"))
                # Execute actual code
                if line2 is not None:
                    for line3 in line2.split("\n"):  # not splitlines!
                        self.more = self.pushline(line3)
                else:
                    self.more = False
                    self._resetbuffer()

        elif ch is self.context._ctrl_code:
            # Read larger block of code (dict)
            msg = self.context._ctrl_code.recv(False)
            if msg:
                # Notify what we're doing
                # (runlargecode() sends on stdin-echo)
                self.context._stat_interpreter.send("Busy")
                self.newPrompt = True
                # Execute code
                self.runlargecode(msg)
                # Reset more stuff
                self._resetbuffer()
                self.more = False

        else:
            # This should not happen, but if it does, just flush!
            ch.recv(False)

    ## Running code in various ways
    # In all cases there is a call for compilecode and a call to execcode

    def _resetbuffer(self):
        """Reset the input buffer."""
        self._buffer = []

    def pushline(self, line):
        """Push a line to the interpreter.

        The line should not have a trailing newline; it may have
        internal newlines.  The line is appended to a buffer and the
        interpreter's _runlines() method is called with the
        concatenated contents of the buffer as source.  If this
        indicates that the command was executed or invalid, the buffer
        is reset; otherwise, the command is incomplete, and the buffer
        is left as it was after the line was appended.  The return
        value is 1 if more input is required, 0 if the line was dealt
        with in some way (this is the same as _runlines()).

        """
        # Get buffer, join to get source
        buffer = self._buffer
        buffer.append(line)
        source = "\n".join(buffer)
        # Clear buffer and run source
        self._resetbuffer()
        more = self._runlines(source, self._filename)
        # Create buffer if needed
        if more:
            self._buffer = buffer
        return more

    def _runlines(self, source, filename="<input>", symbol="single"):
        """Compile and run some source in the interpreter.

        Arguments are as for compile_command().

        One several things can happen:

        1) The input is incorrect; compile_command() raised an
        exception (SyntaxError or OverflowError).  A syntax traceback
        will be printed by calling the showsyntaxerror() method.

        2) The input is incomplete, and more input is required;
        compile_command() returned None.  Nothing happens.

        3) The input is complete; compile_command() returned a code
        object.  The code is executed by calling self.execcode() (which
        also handles run-time exceptions, except for SystemExit).

        The return value is True in case 2, False in the other cases (unless
        an exception is raised).  The return value can be used to
        decide whether to use sys.ps1 or sys.ps2 to prompt the next
        line.

        """

        use_ipython = self._ipython and not self._dbFrames

        # Try compiling.
        # The IPython kernel does not handle incomplete lines, so we check
        # that ourselves ...
        error = None
        try:
            code = self.compilecode(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            error = sys.exc_info()[1]
            code = False

        if use_ipython:
            if code is None:
                # Case 2
                # self._ipython.run_cell('', True)
                return True
            else:
                # Case 1 and 3 handled by IPython
                self._ipython.run_cell(source, True, False)
                return False

        else:
            if code is None:
                # Case 2
                return True
            elif not code:
                # Case 1, a bit awkward way to show the error, but we need
                # to call showsyntaxerror in an exception handler.
                try:
                    raise error
                except Exception:
                    self.showsyntaxerror(filename)
                return False
            else:
                # Case 3
                self.execcode(code)
                return False

    def runlargecode(self, msg, silent=False):
        """To execute larger pieces of code."""

        # Get information
        source, fname, lineno = msg["source"], msg["fname"], msg["lineno"]
        cellName = msg.get("cellName", "")
        source += "\n"

        # Change directory?
        if msg.get("changeDir", False) and os.path.isfile(fname):
            d = os.path.normpath(os.path.normcase(os.path.dirname(fname)))
            if d != os.getcwd():
                os.chdir(d)

        # Construct notification message
        lineno1 = lineno + 1
        lineno2 = lineno + source.count("\n")
        fname_show = fname
        if not fname.startswith("<"):
            fname_show = os.path.split(fname)[1]
        if cellName == fname:
            runtext = '(executing file "%s")\n' % fname_show
        elif cellName:
            runtext = '(executing cell "%s" (line %i of "%s"))\n' % (
                cellName.strip(),
                lineno1,
                fname_show,
            )
            # Try to get the last expression printed in the cell.
            try:
                import ast

                tree = ast.parse(source, fname, "exec")
                if (
                    isinstance(tree.body[-1], ast.Expr)
                    and tree.body[-1].col_offset == 0
                ):
                    e = tree.body[-1]
                    lines = source.splitlines()
                    lines[e.lineno - 1] = "_=\\\n" + lines[e.lineno - 1]
                    source2 = (
                        "\n".join(lines).rstrip()
                        + "\nif _ is not None:\n  print(repr(_))\n"
                    )
                    ast.parse(
                        source2, fname, "exec"
                    )  # This is to make sure it still compiles
                    source = source2
            except Exception:
                pass
        elif lineno1 == lineno2:
            runtext = '(executing line %i of "%s")\n' % (lineno1, fname_show)
        else:
            runtext = '(executing lines %i to %i of "%s")\n' % (
                lineno1,
                lineno2,
                fname_show,
            )
        # Notify IDE
        if not silent:
            self.context._strm_echo.send("\x1b[0;33m%s\x1b[0m" % runtext)
            # Increase counter
            if self._ipython:
                self._ipython.execution_count += 1

        # Bring fname to the canonical form so that pdb recognizes the breakpoints,
        # otherwise filename r"C:\..." would be different from canonical form r"c:\..."
        fname = self.debugger.canonic(fname)

        # Put the line number in the filename (if necessary)
        # Note that we could store the line offset in the _codeCollection,
        # but then we cannot retrieve it for syntax errors.
        if lineno:
            fname = "%s+%i" % (fname, lineno)

        # Try compiling the source
        code = None
        try:
            # Compile
            code = self.compilecode(source, fname, "exec")

        except (OverflowError, SyntaxError, ValueError):
            self.showsyntaxerror(fname)
            return

        if code:
            # Store the source using the (id of the) code object as a key
            self._codeCollection.store_source(code, source)
            # Execute the code
            self.execcode(code)
        else:
            # Incomplete code
            self.write("Could not run code because it is incomplete.\n")

    def runfile(self, fname):
        """To execute the startup script."""

        # Get text (make sure it ends with a newline)
        try:
            encoding = "utf-8-sig"
            with open(fname, "rb") as fd:
                bb = fd.read()
                if PYTHON_VERSION >= 3:
                    fd.seek(0)
                    encoding = tokenize.detect_encoding(fd.readline)[0]
            if PYTHON_VERSION < 3:
                firstline = bb.split(b"\n", 1)[0].decode("ascii", "ignore")
                if firstline.startswith("#") and "coding" in firstline:
                    encoding = firstline.split("coding", 1)[-1].strip(" \t\r\n:=-*")
            source = bb.decode(encoding)
        except Exception:
            printDirect(
                "Could not read script (decoding using %s): %r\n" % (encoding, fname)
            )
            return
        try:
            source = source.replace("\r\n", "\n").replace("\r", "\n")
            if source[-1] != "\n":
                source += "\n"
        except Exception:
            printDirect('Could not execute script: "' + fname + '"\n')
            return

        # Try compiling the source
        code = None
        try:
            # Compile
            code = self.compilecode(source, fname, "exec")
        except (OverflowError, SyntaxError, ValueError):
            time.sleep(0.2)  # Give stdout time to be sent
            self.showsyntaxerror(fname)
            return

        if code:
            # Store the source using the (id of the) code object as a key
            self._codeCollection.store_source(code, source)
            # Execute the code
            self.execcode(code)
        else:
            # Incomplete code
            self.write("Could not run code because it is incomplete.\n")

    def compilecode(self, source, filename, mode, *args, **kwargs):
        """Compile source code.
        Will mangle coding definitions on first two lines.

        * This method should be called with Unicode sources.
        * Source newlines should consist only of LF characters.
        """

        # This method solves pyzo issue 22

        # Split in first two lines and the rest
        parts = source.split("\n", 2)

        # Replace any coding definitions
        ci = "coding is"
        contained_coding = False
        for i in range(len(parts) - 1):
            tmp = parts[i]
            if tmp and tmp[0] == "#" and "coding" in tmp:
                contained_coding = True
                parts[i] = tmp.replace("coding=", ci).replace("coding:", ci)

        # Combine parts again (if necessary)
        if contained_coding:
            source = "\n".join(parts)

        # Convert filename to UTF-8 if Python version < 3
        if PYTHON_VERSION < 3:
            filename = filename.encode("utf-8")

        # Compile
        return self._compile(source, filename, mode, *args, **kwargs)

    def execcode(self, code):
        """Execute a code object.

        When an exception occurs, self.showtraceback() is called to
        display a traceback.  All exceptions are caught except
        SystemExit, which is reraised.

        A note about KeyboardInterrupt: this exception may occur
        elsewhere in this code, and may not always be caught.  The
        caller should be prepared to deal with it.

        The globals variable is used when in debug mode.
        """

        try:
            if self._dbFrames:
                self.apply_breakpoints()
                glob, loc = self.globals, self.locals
            else:
                # Turn debugger on at this point. If there are no breakpoints,
                # the tracing is disabled for better performance.
                self.debugger.set_on()
                self.apply_breakpoints()
                glob, loc = self.globals, self.locals
            exec(code, glob, loc)
        except bdb.BdbQuit:
            self.dbstop_handler()
        except Exception:
            time.sleep(0.2)  # Give stdout some time to send data
            self.showtraceback()
        except KeyboardInterrupt:  # is a BaseException, not an Exception
            time.sleep(0.2)
            self.showtraceback()

    def apply_breakpoints(self):
        """Breakpoints are updated at each time a command is given,
        including commands like "db continue".
        """
        try:
            breaks = self.context._stat_breakpoints.recv()
            self.debugger.clear_all_breaks()
            if breaks:  # Can be None
                for fname in breaks:
                    for linenr in breaks[fname]:
                        self.debugger.set_break(fname, linenr)
            if breaks or self.debugger._last_db_command in ("step", "next", "return"):
                sys.settrace(self.debugger.trace_dispatch)
        except Exception:
            type, value, tb = sys.exc_info()
            del tb
            print("Error while setting breakpoints: %s" % str(value))

    ## Handlers and hooks

    def ipython_pre_run_cell_hook(self, ipython=None):
        """Hook that IPython calls right before executing code."""
        self.debugger.set_on()
        self.apply_breakpoints()

    def ipython_editor_hook(self, ipython, filename, linenum=None, wait=True):
        # Correct line number for cell offset
        filename, linenum = self.correctfilenameandlineno(filename, linenum or 0)
        # Get action string
        if linenum:
            action = "open %i %s" % (linenum, os.path.abspath(filename))
        else:
            action = "open %s" % os.path.abspath(filename)
        # Send
        self.context._strm_action.send(action)

    def ipython_ask_exit(self):
        # Ask the user
        a = input("Do you really want to exit ([y]/n)? ")
        a = a or "y"
        # Close stdin if necessary
        if a.lower() == "y":
            sys.stdin._channel.close()

    def dbstop_handler(self, *args, **kwargs):
        print("Program execution stopped from debugger.")

    ## Writing and error handling

    def write(self, text):
        """Write errors."""
        sys.stderr.write(text)

    def showsyntaxerror(self, filename=None):
        """Display the syntax error that just occurred.
        This doesn't display a stack trace because there isn't one.
        If a filename is given, it is stuffed in the exception instead
        of what was there before (because Python's parser always uses
        "<string>" when reading from a string).

        Pyzo version: support to display the right line number,
        see doc of showtraceback for details.
        """

        # Get info (do not store)
        type_, value, tb = sys.exc_info()
        del tb

        # Work hard to stuff the correct filename in the exception
        if filename and issubclass(type_, SyntaxError):
            try:
                # unpack information
                msg = value.args[0]
                # since Python v3.10 there are also end_lineno and end_offset
                dummy_filename, lineno, offset, line = value.args[1][:4]
                # correct line-number
                fname, lineno = self.correctfilenameandlineno(filename, lineno)
            except Exception:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                value = type_(msg, (fname, lineno, offset, line))
                sys.last_value = value

        # Show syntax error
        strList = traceback.format_exception_only(type_, value)
        for s in strList:
            self.write(s)

    def showtraceback(self, useLastTraceback=False):
        """Display the exception that just occurred.
        We remove the first stack item because it is our own code.
        The output is written by self.write(), below.

        In the pyzo version, before executing a block of code,
        the filename is modified by appending " [x]". Where x is
        the index in a list that we keep, of tuples
        (sourcecode, filename, lineno).

        Here, showing the traceback, we check if we see such [x],
        and if so, we extract the line of code where it went wrong,
        and correct the lineno, so it will point at the right line
        in the editor if part of a file was executed. When the file
        was modified since the part in question was executed, the
        fileno might deviate, but the line of code shown shall
        always be correct...
        """
        # Traceback info:
        # tb_next -> go down the trace
        # tb_frame -> get the stack frame
        # tb_lineno -> where it went wrong
        #
        # Frame info:
        # f_back -> go up (towards caller)
        # f_code -> code object
        # f_locals -> we can execute code here when PM debugging
        # f_globals
        # f_trace -> (can be None) function for debugging? (
        #
        # The traceback module is used to obtain prints from the
        # traceback.

        try:
            if useLastTraceback:
                # Get traceback info from buffered
                etype = sys.last_type
                value = sys.last_value
                tb = sys.last_traceback
            else:
                # Get exception information and remove first, since that's us
                etype, value, tb = sys.exc_info()
                tb = tb.tb_next

                # Store for debugging, but only store if not in debug mode
                if not self._dbFrames:
                    sys.last_type = etype
                    sys.last_value = value
                    sys.last_traceback = tb

            lines = traceback.format_exception(etype, value, tb)

            for line in lines:
                line = self.correctfilenameandlineno_on_line(line)
                self.write(line)

        except Exception:
            etype, value, tb = sys.exc_info()
            tb = None
            t = "An error occured, but then another one when trying to write the traceback: "
            t += str(value) + "\n"
            self.write(t)

    def correctfilenameandlineno(self, fname, lineno):
        """Given a filename and lineno, this function returns
        a modified (if necessary) version of the two.
        As example:
        "foo.py+7", 22  -> "foo.py", 29
        """
        j = fname.rfind("+")
        if j > 0:
            try:
                lineno += int(fname[j + 1 :])
                fname = fname[:j]
            except ValueError:
                pass
        return fname, lineno

    def correctfilenameandlineno_on_line(self, line):
        match = re.search(file_linenr_pattern, line)
        if match:
            i1, i2 = match.start(1), match.end(1)
            i3, i4 = match.start(2), match.end(2)
            offset = int(line[i1:i2].lstrip("+"))
            old_nr = int(line[i3:i4])
            new_nr = old_nr + offset
            return line[:i1] + line[i2:i3] + str(new_nr) + line[i4:]
        return line

    ## Advanced methods

    def switchframe(self, frameindex=0):
        """switches the globals and locals for the shell to the selected frame

        frameindex 0 is the outermost frame

        For normally started shells, this makes no sense, because the interpreter
        already uses the outermost frame, and all inner frames are only the ones of
        the Pyzo kernel.

        But when running an externally started shell, there could be more frames between
        the outermost frame and the first frame of the Pyzo kernel.
        It could be useful to switch to another frame to inspect or modify variables
        there. To switch the globals/locals to the second outermost frame, for example,
        execute the line
            import sys; sys._pyzoInterpreter.switchframe(1)
        in the shell. Note that the switch will be performed only after the execution
        of that line has finished.
        """
        f = sys._getframe()

        frames = [f]
        while f.f_back:
            f = f.f_back
            frames.append(f)
        frames.reverse()

        if self._dbFrames:
            printDirect("cannot switch frame because debugging is active\n")
            return

        try:
            f = frames[frameindex]
        except IndexError:
            printDirect("cannot switch frame because frameindex is invalid\n")
            return

        fname, lineno = self.correctfilenameandlineno(f.f_code.co_filename, f.f_lineno)

        printDirect("setting new base frame for the Pyzo interpreter:\n")
        printDirect(
            '    File "{}", line {}, in {}\n'.format(fname, lineno, f.f_code.co_name)
        )

        self._main_locals = self.locals = f.f_locals
        self._main_globals = self.globals = f.f_globals


class ExecutedSourceCollection:
    """Stores the source of executed pieces of code, so that the right
    traceback can be reproduced when an error occurs. The filename
    (including the +lineno suffix) is used as a key. We monkey-patch
    the linecache module so that we first try our cache to look up the
    lines. In that way we also allow third party modules (e.g. IPython)
    to get the lines for executed cells.
    """

    def __init__(self):
        self._cache = {}
        self._patch()

    def store_source(self, codeObject, source):
        self._cache[codeObject.co_filename] = source

    def _patch(self):
        def getlines(filename, module_globals=None):
            # !!! do not use "import" inside this function because it can
            # cause an infinite recursion loop with the import override in
            # module shiboken6 (PySide6) !!!

            # Try getting the source from our own cache,
            # otherwise fallback to linecache's own cache
            src = self._cache.get(filename, "")
            if src:
                return [line + "\n" for line in src.splitlines()]
            else:
                return linecache._getlines(filename, module_globals)

        # Monkey patch
        if not hasattr(linecache, "_getlines"):
            # Normally, the linecache module is not patched yet.
            # But when starting the Pyzo kernel in an external shell twice in a row,
            # we must not patch the module again to prevent infinite recursion.
            linecache._getlines = linecache.getlines
            linecache.getlines = getlines

        # I hoped this would remove the +lineno for IPython tracebacks,
        # but it doesn't


#         def extract_tb(tb, limit=None):
#             print('aasdasd')
#             import traceback
#             list1 = traceback._extract_tb(tb, limit)
#             list2 = []
#             for (filename, lineno, name, line) in list1:
#                 filename, lineno = sys._pyzoInterpreter.correctfilenameandlineno(filename, lineno)
#                 list2.append((filename, lineno, name, line))
#             return list2
#
#         import traceback
#         traceback._extract_tb = traceback.extract_tb
#         traceback.extract_tb = extract_tb



#!/usr/bin/env python3
"""
Pyzo __main__ module

This module enables starting Pyzo via either "python3 -m pyzo" or
"python3 path/to/pyzo".
"""

import os
import sys
import time
import platform


# Very probably run as a script, either the package or the __main__
# directly. Add parent directory to sys.path and try again.
this_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, os.path.split(this_dir)[0])

import pyzo


TESTING = "--test" in sys.argv


def write(*msg):
    print(*msg)
    if os.getenv("PYZO_LOG", ""):
        with open(os.getenv("PYZO_LOG"), "at") as f:
            f.write(" ".join(msg) + "\n")


def main():
    if TESTING:
        dt = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        write(f"Testing Pyzo source ({dt} UTC)")
        write(platform.platform())
        write(sys.version)

    write(f"Pyzo {pyzo.__version__}")
    pyzo.start()
    write("Stopped")  # may be written to log twice because Pyzo defers stdout


if __name__ == "__main__":
    main()



import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets
from pyzo import translate

tool_name = translate("pyzoExpressionViewer", "Expression viewer")
tool_summary = "View values of expressions."


class ExpressionProxy(QtCore.QObject):
    """ExpressionProxy

    A proxy class to handle the asynchonous behaviour of getting information
    from the shell. The expression viewer tool asks for certain expressions,
    and this class notifies when new data is available using a qt signal.
    """

    haveNewData = QtCore.Signal()

    def __init__(self):
        super().__init__()

        self._expressions = []
        self._response = []

        # Bind to events
        pyzo.shells.currentShellChanged.connect(self._onCurrentShellChanged)
        pyzo.shells.currentShellStateChanged.connect(self._onCurrentShellStateChanged)

    def setExpressions(self, expressions):
        """Set the expressions to be evaluated."""
        self._expressions = list(expressions)

    def getExpressions(self):
        return self._expressions[:]

    def evaluateExpressions(self):
        """trigger evaluation of the expressions

        when finished, signal "haveNewData" will be emitted
        """
        shell = pyzo.shells.getCurrentShell()
        if self._expressions and shell and shell._state not in ("Busy", "Very busy"):
            future = shell._request.evalMultiple(self._expressions)
            future.add_done_callback(self._processResponse)

    def _onCurrentShellChanged(self):
        """When no shell is selected now, update this. In all other cases,
        method _onCurrentShellStateChanged will be fired too.
        """
        shell = pyzo.shells.getCurrentShell()
        if not shell:
            self._response = []
            self.haveNewData.emit()

    def _onCurrentShellStateChanged(self):
        self.evaluateExpressions()

    def _processResponse(self, future):
        """We got a response, save the result and notify the tree."""

        self._response = []

        # Process future
        if future.cancelled():
            pass  # no living kernel
        elif future.exception():
            print("Introspect exception:", future.exception())
        else:
            self._response = future.result()

        self.haveNewData.emit()


class ExpressionManager:
    """This is the single source of truth where all expressions and their values are stored."""

    def __init__(self):
        self._expressions = {}  # key: expression, value: (enabled, stringvalue)

    def _cleanUpExpressionString(self, expression):
        valid = True
        expression = expression.strip()
        if len(expression) == 0:
            valid = False
            print("WARNING: expression is empty")
        elif len(expression.splitlines()) != 1:
            # using len splitlines because checking for \r and \n is not enough:
            #   'xx\u2028yy'.splitlines() --> ['xx', 'yy']
            valid = False
            print(
                "WARNING: expression",
                repr(expression),
                "contains new-line like characters",
            )
        elif expression.startswith("#"):
            valid = False
            print(
                "WARNING: expression",
                repr(expression),
                "starts with a comment character",
            )
        return valid, expression

    def clear(self):
        self._expressions.clear()

    def hasExpression(self, expression):
        return expression in self._expressions

    def getExpressions(self):
        """returns the expressions as a dict

        key: expression
        value: (enabled, stringvalue)
        """
        return self._expressions.copy()

    def addExpression(self, expression, enabled=True, value="[not evaluated yet]"):
        valid, expression = self._cleanUpExpressionString(expression)
        if valid:
            self._expressions[expression.strip()] = (enabled, value)
        success = valid
        return success

    def modifyExpression(
        self, expression, newExpression=None, newEnabled=None, newValue=None
    ):
        success = True
        if newExpression is None:
            if expression in self._expressions:
                enabled, stringvalue = self._expressions[expression]
                if newEnabled is None:
                    newEnabled = enabled
                if newValue is None:
                    newValue = stringvalue
                self._expressions[expression] = (bool(newEnabled), newValue)
        elif newExpression != expression:
            # modify the expression, and keep the order of all expressions in the dict
            oldExpressions = self._expressions.copy()
            self.clear()
            for currentExpression, (enabled, stringvalue) in oldExpressions.items():
                if expression == currentExpression:
                    if newEnabled is None:
                        newEnabled = enabled
                    if newValue is None:
                        newValue = stringvalue
                    if not self.addExpression(newExpression, newEnabled):
                        # update failed --> keep old value
                        self.addExpression(expression, enabled)
                        success = False
                else:
                    self._expressions[currentExpression] = (enabled, stringvalue)
        return success

    def toString(self):
        """converts the expressions to a multi-line string

        Each line contains the expression, and if not enabled, it will be prefixed by "# ".
        The stringvalue (result from evaluating the expression) is ignored.
        """
        ll = []
        for expression, (enabled, stringvalue) in self._expressions.items():
            prefix = "" if enabled else "# "
            ll.append(prefix + expression)
        return "\n".join(ll)

    def fromString(self, text):
        """extracts expressions from a multi-line string

        Each line represents an expression.
        If the line is prefixed by a "#", the expression will be marked as disabled.
        """
        oldExpressions = self._expressions.copy()
        self.clear()
        for line in text.splitlines():
            enabled = True
            line = line.strip()
            while line.startswith("#"):
                enabled = False
                line = line.lstrip("#")
                line = line.lstrip()  # this will also strip non-ascii whitespace
            expression = line
            _, value = oldExpressions.get(expression, (False, "[not evaluated yet]"))
            self.addExpression(expression, enabled, value)


class PyzoExpressionViewer(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        # create proxy
        self._proxy = ExpressionProxy()
        self._proxy.haveNewData.connect(self._processResults)

        # create expression manager
        self._em = ExpressionManager()

        # create tool buttons
        self._btnRefresh = QtWidgets.QToolButton(self)
        self._btnRefresh.setIcon(pyzo.icons.arrow_refresh)
        self._btnRefresh.setIconSize(QtCore.QSize(16, 16))
        self._btnRefresh.setToolTip(
            "manually refresh expressions (e.g. for code running in the event loop)\n"
            "this will also refresh the Workspace tool, if open"
        )
        self._btnRefresh.pressed.connect(self._onButtonRefresh)

        # create edit toggle button
        self._btnEdit = QtWidgets.QToolButton(self)
        self._btnEdit.setIcon(pyzo.icons.application_edit)
        self._btnEdit.setIconSize(QtCore.QSize(16, 16))
        self._btnEdit.setCheckable(True)
        self._btnEdit.setToolTip("toggle edit mode")
        self._btnEdit.toggled.connect(self._onButtonEditToggled)

        # create add-from-editor button
        self._btnAddFromEditor = QtWidgets.QToolButton(self)
        self._btnAddFromEditor.setIcon(pyzo.icons.add)
        self._btnAddFromEditor.setIconSize(QtCore.QSize(16, 16))
        self._btnAddFromEditor.setToolTip("add selected expression from current editor")
        self._btnAddFromEditor.pressed.connect(self._onButtonAddFromEditor)

        # create clear button
        self._btnRemoveAll = QtWidgets.QToolButton(self)
        self._btnRemoveAll.setIcon(pyzo.icons.application_eraser)
        self._btnRemoveAll.setIconSize(QtCore.QSize(16, 16))
        self._btnRemoveAll.setToolTip("remove all expressions")
        self._btnRemoveAll.pressed.connect(self._onButtonRemoveAll)

        # create monospaced text input
        self._textEdit = QtWidgets.QPlainTextEdit()
        font = self._textEdit.font()
        font.setFamily(pyzo.config.view.fontname)
        self._textEdit.setFont(font)

        # set the top bar layout
        self._topBarLayout = QtWidgets.QHBoxLayout()
        self._topBarLayout.addWidget(self._btnRefresh)
        self._topBarLayout.addWidget(self._btnEdit)
        self._topBarLayout.addWidget(self._btnAddFromEditor)
        self._topBarLayout.addWidget(self._btnRemoveAll)
        self._topBarLayout.addStretch()

        # create the tree widget (but only used as a multi-column list)
        tree = self._tree = QtWidgets.QTreeWidget(self)

        # set header stuff
        tree.setHeaderHidden(False)
        tree.setColumnCount(2)
        tree.setHeaderLabels(["Expression", "Result"])
        tree.header().setStretchLastSection(True)

        # setup context menu for the tree widget
        tree.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        tree.customContextMenuRequested.connect(self._onTreeContextMenuRequested)
        self._treeContextMenu = QtWidgets.QMenu()
        self._treeContextMenu.triggered.connect(self._onTreeContextMenuTriggered)

        # nice rows
        tree.setAlternatingRowColors(True)
        tree.setRootIsDecorated(False)

        tree.itemChanged.connect(self._onTreeItemChanged)

        # set the main layout
        mainLayout = QtWidgets.QVBoxLayout(self)
        mainLayout.setSpacing(2)
        mainLayout.addLayout(self._topBarLayout)
        mainLayout.addWidget(self._tree)
        mainLayout.addWidget(self._textEdit)

        # set margins
        margin = pyzo.config.view.widgetMargin
        mainLayout.setContentsMargins(margin, margin, margin, margin)

        self.setLayout(mainLayout)

        # add example entries
        self._em.addExpression("1 + 1", False)
        self._em.addExpression("len(dir())", False)
        self._em.addExpression("repr('\\n'.join(globals())[:30])", False)
        self._textEdit.setPlainText(self._em.toString())

        self._setEditMode(False)

    def _onTreeContextMenuRequested(self, pos):
        """Show the context menu."""

        # get the selected item
        item = self._tree.currentItem()
        if not item:
            return

        # create menu
        self._treeContextMenu.clear()
        commands = [
            ("ShowInEditMode", "Show in edit mode"),
            ("ToggleRepr", "Toggle repr"),
            ("EvaluateInShell", "Evaluate in shell"),
        ]
        for request, display in commands:
            action = self._treeContextMenu.addAction(display)
            action._what = request
            action._item = item

        # show the context menu
        self._treeContextMenu.popup(QtGui.QCursor.pos())

    def _onTreeContextMenuTriggered(self, action):
        """Process a request from the context menu."""
        self._performTreeItemCommand(action._what, action._item)

    def _performTreeItemCommand(self, cmd, item):
        assert not self._inEditMode()
        expr = item.text(0)
        if cmd == "ToggleRepr":
            prefix, postfix = "repr(", ")"
            if expr.startswith(prefix) and expr.endswith(postfix):
                expr = expr[len(prefix) : len(expr) - len(postfix)]
            else:
                expr = prefix + expr + postfix
            item.setText(0, expr)
        elif cmd == "ShowInEditMode":
            # select the whole corresponding line in the text widget
            ind = self._tree.indexOfTopLevelItem(item)  # -1 if not found
            self._btnEdit.toggle()
            cursor = self._textEdit.textCursor()
            cursor.movePosition(cursor.MoveOperation.Start)
            cursor.movePosition(
                cursor.MoveOperation.NextBlock,
                cursor.MoveMode.MoveAnchor,
                ind,
            )
            cursor.movePosition(
                cursor.MoveOperation.EndOfBlock,
                cursor.MoveMode.KeepAnchor,
            )
            self._textEdit.setTextCursor(cursor)
        elif cmd == "EvaluateInShell":
            shell = pyzo.shells.getCurrentShell()
            if shell is not None:
                pyzo.command_history.append(expr)
                shell.executeCommand(expr + "\n")

    def addExpression(self, expression):
        if self._inEditMode():
            text = self._textEdit.toPlainText()
            if expression not in text.splitlines():
                if text != "" and not text.endswith("\n"):
                    text += "\n"
                text += expression + "\n"
                self._textEdit.setPlainText(text)
        else:
            self._em.addExpression(expression, True)
            self._updateTree()
            self._sendExpressionsToProxy()

    def _inEditMode(self):
        return self._btnEdit.isChecked()

    def _setEditMode(self, enabled):
        self._btnRefresh.setEnabled(not enabled)
        if enabled:
            self._proxy.setExpressions([])
            text = self._em.toString()
            if len(text) > 0:
                text += "\n"
            self._textEdit.setPlainText(text)
            self._tree.hide()

            # move to the empty last line -- for immediately entering a new expression
            cursor = self._textEdit.textCursor()
            cursor.movePosition(cursor.MoveOperation.End)
            self._textEdit.setTextCursor(cursor)

            self._textEdit.show()
            self._textEdit.setFocus()
        else:
            self._em.fromString(self._textEdit.toPlainText())
            self._updateTree()
            self._sendExpressionsToProxy()
            self._proxy.evaluateExpressions()
            self._textEdit.hide()
            self._tree.show()
            self._tree.setFocus()

    def _sendExpressionsToProxy(self):
        self._proxy.setExpressions(
            [
                expr
                for expr, (enabled, v) in self._em.getExpressions().items()
                if enabled
            ]
        )

    def _onButtonEditToggled(self, enabled):
        self._setEditMode(enabled)

    def _onButtonRefresh(self):
        self.manualRefresh()
        ws = pyzo.toolManager.getTool("pyzoworkspace")
        if ws is not None:
            ws.manualRefresh()

    def _onButtonAddFromEditor(self):
        ed = pyzo.editors.getCurrentEditor()
        if ed is not None:
            expr = ed.textCursor().selectedText().strip()
            self.addExpression(expr)
            self.manualRefresh()

    def _onButtonRemoveAll(self):
        self._em.clear()
        self._sendExpressionsToProxy()
        if self._inEditMode():
            self._textEdit.setPlainText("")
        else:
            self._updateTree()

    def manualRefresh(self):
        """manually refresh the expressions"""
        self._proxy.evaluateExpressions()

    def _onTreeItemChanged(self, item, column):
        self._tree.blockSignals(True)
        try:
            if column == 1:
                # restore original value (tool tip text is the same as the original value)
                item.setText(1, item.toolTip(1))
            elif column == 0:
                if item.text(0) != item.toolTip(0):
                    # update the expression
                    origExpr = item.toolTip(0)
                    newExpr = item.text(0)
                    if not self._em.modifyExpression(origExpr, newExpression=newExpr):
                        item.setText(0, origExpr)
                    else:
                        item.setToolTip(0, newExpr)

                # update the enabled status, no matter if it was changed or not
                enabled = item.checkState(0) == QtCore.Qt.CheckState.Checked
                self._em.modifyExpression(item.text(0), newEnabled=enabled)

                self._sendExpressionsToProxy()
                self._proxy.evaluateExpressions()
        finally:
            self._tree.blockSignals(False)

    def _updateTree(self):
        tree = self._tree
        CS = QtCore.Qt.CheckState

        tree.blockSignals(True)
        try:
            # The widget is not cleared to keep its current scroll bar position.
            i = -1
            for expr, (enabled, stringvalue) in self._em.getExpressions().items():
                i += 1
                item = tree.topLevelItem(i)
                if item is None:
                    # Create item
                    item = QtWidgets.QTreeWidgetItem((expr, stringvalue))
                    tree.addTopLevelItem(item)
                    item.setFlags(item.flags() | QtCore.Qt.ItemFlag.ItemIsEditable)
                else:
                    item.setText(0, expr)
                    item.setText(1, stringvalue)
                    item.setSelected(False)

                item.setCheckState(0, CS.Checked if enabled else CS.Unchecked)
                item.setToolTip(0, expr)
                item.setToolTip(1, stringvalue)

            for _ in range(i + 1, tree.topLevelItemCount()):
                tree.takeTopLevelItem(i + 1)  # delete remaining entries

        finally:
            tree.blockSignals(False)

    def _processResults(self):
        # assign the results to the corresponding expressions in the expression manager
        for expression, success, stringvalue in self._proxy._response:
            if self._em.hasExpression(expression):
                if not success:
                    stringvalue = "[ERROR: " + stringvalue + "]"
                self._em.modifyExpression(expression, newValue=stringvalue)

        if not self._inEditMode():
            self._updateTree()

    def keyPressEvent(self, event):
        # leave "edit" mode if ESC key is pressed
        if event.key() == QtCore.Qt.Key.Key_Escape and not event.modifiers():
            if self._inEditMode():
                self._btnEdit.click()
                return
        super().keyPressEvent(event)
